# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the behaviour classes of the tac participation skill."""

import logging
from pathlib import Path
from typing import cast
from unittest.mock import patch

import pytest

from aea.helpers.transaction.base import Terms
from aea.protocols.dialogue.base import DialogueMessage
from aea.test_tools.test_skill import BaseSkillTestCase, COUNTERPARTY_ADDRESS

from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.protocols.tac.message import TacMessage
from packages.fetchai.skills.tac_participation.behaviours import (
    TacSearchBehaviour,
    TransactionProcessBehaviour,
)
from packages.fetchai.skills.tac_participation.dialogues import TacDialogues
from packages.fetchai.skills.tac_participation.game import Game, Phase

from tests.conftest import ROOT_DIR


class TestTacSearchBehaviour(BaseSkillTestCase):
    """Test tac behaviour of tac participation."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_participation")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.tac_search_behaviour = cast(
            TacSearchBehaviour, cls._skill.skill_context.behaviours.tac_search
        )
        cls.game = cast(Game, cls._skill.skill_context.game)
        cls.logger = cls.tac_search_behaviour.context.logger

    def test_setup(self):
        """Test the setup method of the tac_search behaviour."""
        assert self.tac_search_behaviour.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_act_i(self):
        """Test the act method of the tac_search behaviour where phase is not PRE_GAME."""
        # setup
        self.game._phase = Phase.GAME

        # operation
        with patch.object(self.logger, "log"):
            self.tac_search_behaviour.act()

        # after
        assert self.game.phase == Phase.GAME
        self.assert_quantity_in_outbox(0)

    def test_act_ii(self):
        """Test the act method of the tac_search behaviour where phase is PRE_GAME."""
        # setup
        self.game._phase = Phase.PRE_GAME
        mocked_query = "some_query"

        # operation
        with patch.object(self.game, "get_game_query", return_value=mocked_query):
            with patch.object(self.logger, "log") as mock_logger:
                self.tac_search_behaviour.act()

        # after
        self.assert_quantity_in_outbox(1)

        # _search_for_tac
        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=OefSearchMessage,
            performative=OefSearchMessage.Performative.SEARCH_SERVICES,
            to=self.skill.skill_context.search_service_address,
            sender=self.skill.skill_context.agent_address
            + "_"
            + str(self.skill.skill_context.skill_id),
            query=mocked_query,
        )
        assert has_attributes, error_str

        mock_logger.assert_any_call(
            logging.INFO, f"searching for TAC, search_id={message.dialogue_reference}"
        )

    def test_teardown(self):
        """Test the teardown method of the tac_search behaviour."""
        assert self.tac_search_behaviour.teardown() is None
        self.assert_quantity_in_outbox(0)


class TestTransactionProcessBehaviour(BaseSkillTestCase):
    """Test tac behaviour of tac participation."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_participation")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.transaction_process_behaviour = cast(
            TransactionProcessBehaviour,
            cls._skill.skill_context.behaviours.transaction_processing,
        )
        cls.game = cast(Game, cls._skill.skill_context.game)
        cls.logger = cls.transaction_process_behaviour.context.logger

        cls.tac_dialogues = cast(TacDialogues, cls._skill.skill_context.tac_dialogues)

        cls.list_of_tac_messages = (
            DialogueMessage(
                TacMessage.Performative.REGISTER, {"agent_name": "some_agent_name"}
            ),
            DialogueMessage(
                TacMessage.Performative.GAME_DATA,
                {
                    "amount_by_currency_id": {"FET": 1},
                    "exchange_params_by_currency_id": {"FET": 1.0},
                    "quantities_by_good_id": {"2": 10},
                    "utility_params_by_good_id": {"2": 1.0},
                    "fee_by_currency_id": {"1": 1},
                    "agent_addr_to_name": {COUNTERPARTY_ADDRESS: "some_agent_name"},
                    "currency_id_to_name": {"1": "FETCH"},
                    "good_id_to_name": {"2": "Good_1"},
                    "version_id": "v1",
                },
            ),
        )

        cls.tx_ids = ["tx_1", "tx_2"]
        cls.terms_1 = Terms(
            "some_ledger_id",
            "some_sender_address_1",
            "some_counterparty_address_1",
            {"1": 10},
            {"2": 5},
            "some_nonce_1",
            fee_by_currency_id={"1": 1},
        )
        cls.terms_2 = Terms(
            "some_ledger_id",
            "some_sender_address_2",
            "some_counterparty_address_2",
            {"1": 11},
            {"2": 6},
            "some_nonce_2",
            fee_by_currency_id={"1": 2},
        )
        cls.terms = [cls.terms_1, cls.terms_2]
        cls.sender_signatures = ["sender_signature_1", "sender_signature_2"]
        cls.counterparty_signatures = [
            "counterparty_signature_1",
            "counterparty_signature_2",
        ]

        cls.txs = {
            cls.tx_ids[0]: {
                "terms": cls.terms[0],
                "sender_signature": cls.sender_signatures[0],
                "counterparty_signature": cls.counterparty_signatures[0],
            },
            cls.tx_ids[1]: {
                "terms": cls.terms[1],
                "sender_signature": cls.sender_signatures[1],
                "counterparty_signature": cls.counterparty_signatures[1],
            },
        }

    def test_setup(self):
        """Test the setup method of the transaction_process behaviour."""
        assert self.transaction_process_behaviour.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_act_i(self):
        """Test the act method of the transaction_process behaviour where phase is not GAME."""
        # setup
        self.game._phase = Phase.PRE_GAME

        # operation
        with patch.object(self.logger, "log"):
            self.transaction_process_behaviour.act()

        # after
        assert self.game.phase == Phase.PRE_GAME
        self.assert_quantity_in_outbox(0)

    def test_act_ii(self):
        """Test the act method of the transaction_process behaviour where phase is GAME."""
        # setup
        self.game._phase = Phase.GAME
        no_tx = len(self.txs)
        self.skill.skill_context._agent_context._shared_state = {
            "transactions": self.txs
        }

        tac_dialogue = self.prepare_skill_dialogue(
            self.tac_dialogues, self.list_of_tac_messages,
        )
        self.game._tac_dialogue = tac_dialogue

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.transaction_process_behaviour.act()

        # after
        self.assert_quantity_in_outbox(no_tx)

        # _process_transactions
        count = 0
        while count != no_tx:
            mock_logger.assert_any_call(
                logging.INFO, f"sending transaction {self.tx_ids[count]} to controller."
            )
            message = self.get_message_from_outbox()
            has_attributes, error_str = self.message_has_attributes(
                actual_message=message,
                message_type=TacMessage,
                performative=TacMessage.Performative.TRANSACTION,
                to=COUNTERPARTY_ADDRESS,
                sender=self.skill.skill_context.agent_address,
                target=tac_dialogue.last_message.message_id - no_tx + count,
                transaction_id=self.tx_ids[count],
                ledger_id=self.terms[count].ledger_id,
                sender_address=self.terms[count].sender_address,
                counterparty_address=self.terms[count].counterparty_address,
                amount_by_currency_id=self.terms[count].amount_by_currency_id,
                fee_by_currency_id=self.terms[count].fee_by_currency_id,
                quantities_by_good_id=self.terms[count].quantities_by_good_id,
                sender_signature=self.sender_signatures[count],
                counterparty_signature=self.counterparty_signatures[count],
                nonce=self.terms[count].nonce,
            )
            assert has_attributes, error_str
            count += 1

    def test_process_transactions_tac_dialogue_is_empty(self):
        """Test the _process_transactions method of the transaction_process behaviour where last message of tac_dialogue is None."""
        # setup
        self.game._phase = Phase.GAME
        self.skill.skill_context._agent_context._shared_state = {
            "transactions": self.txs
        }

        tac_dialogue = self.prepare_skill_dialogue(
            self.tac_dialogues, self.list_of_tac_messages,
        )

        tac_dialogue._incoming_messages = []
        tac_dialogue._outgoing_messages = []

        self.game._tac_dialogue = tac_dialogue

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with pytest.raises(ValueError, match="No last message available."):
                self.transaction_process_behaviour.act()

        # after
        mock_logger.assert_any_call(
            logging.INFO, f"sending transaction {self.tx_ids[0]} to controller."
        )

    def test_process_transactions_invalid_tx(self):
        """Test the _process_transactions method of the transaction_process behaviour where transactions are None."""
        # setup
        self.game._phase = Phase.GAME
        self.skill.skill_context._agent_context._shared_state = {
            "transactions": {self.tx_ids[0]: None, self.tx_ids[1]: None}
        }

        tac_dialogue = self.prepare_skill_dialogue(
            self.tac_dialogues, self.list_of_tac_messages,
        )
        self.game._tac_dialogue = tac_dialogue

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with pytest.raises(
                ValueError, match=f"Tx for id={self.tx_ids[0]} not found."
            ):
                self.transaction_process_behaviour.act()

        # after
        mock_logger.assert_any_call(
            logging.INFO, f"sending transaction {self.tx_ids[0]} to controller."
        )

    def test_teardown(self):
        """Test the teardown method of the transaction_process behaviour."""
        assert self.transaction_process_behaviour.teardown() is None
        self.assert_quantity_in_outbox(0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the helpers module of the tac control skill."""

from pathlib import Path
from unittest.mock import patch

import pytest

from aea.exceptions import AEAEnforceError
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.skills.tac_control_contract.helpers import (
    ERC1155Contract,
    _sample_good_instances,
    determine_scaling_factor,
    generate_currency_endowments,
    generate_currency_id_to_name,
    generate_currency_ids,
    generate_equilibrium_prices_and_holdings,
    generate_exchange_params,
    generate_good_endowments,
    generate_good_id_to_name,
    generate_good_ids,
    generate_utility_params,
)

from tests.conftest import ROOT_DIR


class TestHelpers(BaseSkillTestCase):
    """Test Helper module methods of tac control."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_control")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()

    def test_generate_good_ids_succeeds(self):
        """Test the generate_good_ids of Helpers module which succeeds."""
        expected_list = [1, 2, 3, 4, 5]
        with patch.object(
            ERC1155Contract, "generate_token_ids", return_value=expected_list
        ):
            good_ids = generate_good_ids(5)
        assert good_ids == expected_list

    def test_generate_good_ids_fails(self):
        """Test the generate_good_ids of Helpers module which fails because generate_token_ids generates wrong good ids."""
        expected_list = [1, 2, 3, 4, 5, 6]
        with patch.object(
            ERC1155Contract, "generate_token_ids", return_value=expected_list
        ):
            with pytest.raises(
                AEAEnforceError,
                match="Length of good ids and number of goods must match.",
            ):
                assert generate_good_ids(5)

    def test_generate_currency_ids_succeeds(self):
        """Test the generate_good_ids of Helpers module which succeeds."""
        expected_list = [1, 2, 3, 4, 5]
        with patch.object(
            ERC1155Contract, "generate_token_ids", return_value=expected_list
        ):
            currency_ids = generate_currency_ids(5)
        assert currency_ids == expected_list

    def test_generate_currency_ids_fails(self):
        """Test the generate_good_ids of Helpers module which fails because generate_token_ids generates wrong currency ids."""
        expected_list = [1, 2, 3, 4, 5, 6]
        with patch.object(
            ERC1155Contract, "generate_token_ids", return_value=expected_list
        ):
            with pytest.raises(
                AEAEnforceError,
                match="Length of currency ids and number of currencies must match.",
            ):
                assert generate_currency_ids(5)

    def test_generate_currency_id_to_name(self):
        """Test the generate_currency_id_to_name of Helpers module."""
        expected_currency_id_to_name = {
            "1": "FT_1",
            "3": "FT_3",
            "5": "FT_5",
            "7": "FT_7",
            "9": "FT_9",
        }
        currency_id_to_name = generate_currency_id_to_name([1, 3, 5, 7, 9])
        assert currency_id_to_name == expected_currency_id_to_name

    def test_generate_good_id_to_name(self):
        """Test the generate_good_id_to_name of Helpers module."""
        expected_good_id_to_name = {
            "1": "FT_1",
            "3": "FT_3",
            "5": "FT_5",
            "7": "FT_7",
            "9": "FT_9",
        }
        good_id_to_name = generate_good_id_to_name([1, 3, 5, 7, 9])
        assert good_id_to_name == expected_good_id_to_name

    def test_determine_scaling_factor(self):
        """Test the determine_scaling_factor of Helpers module."""
        money_endowment = 53730411
        scaling_factor = determine_scaling_factor(money_endowment)
        assert scaling_factor == 10000000.0

    def test_generate_good_endowments(self):
        """Test the generate_good_endowments of Helpers module."""
        endowments = generate_good_endowments(
            ["ag_1_add", "ag_2_add"], ["good_id_1", "good_id_2"], 2, 1, 1
        )
        assert "good_id_1" in endowments["ag_1_add"]
        assert "good_id_2" in endowments["ag_1_add"]
        assert "good_id_1" in endowments["ag_2_add"]
        assert "good_id_2" in endowments["ag_2_add"]

    def test_generate_utility_params(self):
        """Test the generate_utility_params of Helpers module."""
        utility_function_params = generate_utility_params(
            ["ag_1_add", "ag_2_add"], ["good_id_1", "good_id_2"], 1000.0
        )
        assert "good_id_1" in utility_function_params["ag_1_add"].keys()
        assert "good_id_2" in utility_function_params["ag_1_add"].keys()
        assert "good_id_1" in utility_function_params["ag_2_add"].keys()
        assert "good_id_2" in utility_function_params["ag_2_add"].keys()

    def test_sample_good_instances(self):
        """Test the _sample_good_instances of Helpers module."""
        nb_instances = _sample_good_instances(2, ["good_id_1", "good_id_2"], 2, 1, 1)
        assert type(nb_instances["good_id_1"]) == int
        assert type(nb_instances["good_id_2"]) == int

    def test_generate_currency_endowments(self):
        """Test the generate_currency_endowments of Helpers module."""
        currency_endowments = generate_currency_endowments(
            ["ag_1_add", "ag_2_add"], ["currency_id_1", "currency_id_2"], 10
        )
        assert "currency_id_1" in currency_endowments["ag_1_add"].keys()
        assert "currency_id_2" in currency_endowments["ag_2_add"].keys()
        assert currency_endowments["ag_1_add"]["currency_id_1"] == 10
        assert currency_endowments["ag_1_add"]["currency_id_2"] == 10

        assert currency_endowments["ag_2_add"]["currency_id_1"] == 10
        assert currency_endowments["ag_2_add"]["currency_id_2"] == 10

    def test_generate_exchange_params(self):
        """Test the generate_exchange_params of Helpers module."""
        currency_endowments = generate_exchange_params(
            ["ag_1_add", "ag_2_add"], ["currency_id_1", "currency_id_2"]
        )
        assert "currency_id_1" in currency_endowments["ag_1_add"].keys()
        assert "currency_id_2" in currency_endowments["ag_2_add"].keys()
        assert currency_endowments["ag_1_add"]["currency_id_1"] == 1.0
        assert currency_endowments["ag_1_add"]["currency_id_2"] == 1.0

        assert currency_endowments["ag_2_add"]["currency_id_1"] == 1.0
        assert currency_endowments["ag_2_add"]["currency_id_2"] == 1.0

    def test_generate_equilibrium_prices_and_holdings(self):
        """Test the generate_equilibrium_prices_and_holdings of Helpers module."""
        (
            eq_prices_dict,
            eq_good_holdings_dict,
            eq_currency_holdings_dict,
        ) = generate_equilibrium_prices_and_holdings(
            {"ag_1": {"good_1": 1}},
            {"ag_1": {"good_1": 1.0}},
            {"ag_1": {"currency_1": 10}},
            {"ag_1": {"currency_1": 1.0}},
            2.0,
        )

        assert len(eq_prices_dict) == 1
        assert type(eq_prices_dict["good_1"]) == float

        assert len(eq_good_holdings_dict) == 1
        assert len(eq_good_holdings_dict["ag_1"]) == 1
        assert type(eq_good_holdings_dict["ag_1"]["good_1"]) == float

        assert len(eq_currency_holdings_dict) == 1
        assert type(eq_currency_holdings_dict["ag_1"]) == float



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for the p2p_stub connection."""

import asyncio
import os
import shutil
import tempfile
from pathlib import Path

import pytest

from aea.configurations.base import ConnectionConfig
from aea.identity.base import Identity
from aea.mail.base import Envelope

from packages.fetchai.connections.p2p_stub.connection import P2PStubConnection
from packages.fetchai.protocols.default.message import DefaultMessage


SEPARATOR = ","


def make_test_envelope(to_="any", sender_="sender") -> Envelope:
    """Create a test envelope."""
    msg = DefaultMessage(
        dialogue_reference=("", ""),
        message_id=1,
        target=0,
        performative=DefaultMessage.Performative.BYTES,
        content=b"hello",
    )
    msg.to = to_
    msg.sender = sender_
    envelope = Envelope(
        to=to_, sender=sender_, protocol_id=DefaultMessage.protocol_id, message=msg,
    )
    return envelope


class Testp2pStubConnectionReception:
    """Test that the stub connection is implemented correctly."""

    def setup(self):
        """Set the test up."""
        self.cwd = os.getcwd()
        self.tmpdir = Path(tempfile.mkdtemp())
        d = self.tmpdir / "test_p2p_stub"
        d.mkdir(parents=True)

        configuration = ConnectionConfig(
            namespace_dir=d, connection_id=P2PStubConnection.connection_id,
        )
        self.loop = asyncio.get_event_loop()
        self.identity1 = Identity("test", "con1")
        self.identity2 = Identity("test", "con2")
        self.connection1 = P2PStubConnection(
            configuration=configuration, identity=self.identity1
        )
        self.connection2 = P2PStubConnection(
            configuration=configuration, identity=self.identity2
        )
        os.chdir(self.tmpdir)

    @pytest.mark.asyncio
    async def test_send(self):
        """Test that the connection receives what has been enqueued in the input file."""
        await self.connection1.connect()
        assert self.connection1.is_connected

        await self.connection2.connect()
        assert self.connection2.is_connected

        envelope = make_test_envelope(to_="con2")
        await self.connection1.send(envelope)

        received_envelope = await asyncio.wait_for(
            self.connection2.receive(), timeout=5
        )
        assert received_envelope
        assert received_envelope.message == envelope.message.encode()

    def teardown(self):
        """Clean up after tests."""
        os.chdir(self.cwd)
        self.loop.run_until_complete(self.connection1.disconnect())
        self.loop.run_until_complete(self.connection2.disconnect())
        try:
            shutil.rmtree(self.tmpdir)
        except (OSError, IOError):
            pass



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the parameters module of the tac control skill."""

import datetime
import logging
from pathlib import Path
from unittest.mock import Mock, patch

import pytest

from aea.exceptions import AEAEnforceError
from aea.helpers.search.models import Location
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.skills.tac_control.parameters import Parameters

from tests.conftest import ROOT_DIR


class TestParameters(BaseSkillTestCase):
    """Test Parameters module of tac control."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_control")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.kwargs = {
            "ledger_id": "some_ledger_id",
            "contract_address": None,
            "good_ids": [],
            "currency_ids": [],
            "min_nb_agents": 2,
            "money_endowment": 200,
            "nb_goods": 9,
            "nb_currencies": 1,
            "tx_fee": 1,
            "base_good_endowment": 2,
            "lower_bound_factor": 1,
            "upper_bound_factor": 1,
            "registration_start_time": "01 01 2020  00:01",
            "registration_timeout": 60,
            "item_setup_timeout": 60,
            "competition_timeout": 300,
            "inactivity_timeout": 30,
            "whitelist": [],
            "location": {"longitude": 0.1270, "latitude": 51.5194},
            "service_data": {"key": "tac", "value": "v1"},
            "name": "parameters",
            "skill_context": cls._skill.skill_context,
        }
        cls.parameters = Parameters(**cls.kwargs)

    @staticmethod
    def _time(date_time: str):
        return datetime.datetime.strptime(date_time, "%d %m %Y %H:%M")

    def test_init_now_after_start(self):
        """Test the init of Parameters where now is after the registration_start_time."""
        mocked_start_time = self._time("01 01 2020 00:01")
        mocked_now_time = self._time("01 01 2020 00:02")

        datetime_mock = Mock(wraps=datetime.datetime)
        datetime_mock.now.return_value = mocked_now_time

        with patch("datetime.datetime", new=datetime_mock):
            with patch.object(self.parameters.context.logger, "log") as mocked_logger:
                Parameters(**self.kwargs)
        mocked_logger.assert_any_call(
            logging.WARNING,
            f"TAC registration start time {mocked_start_time} is in the past! Deregistering skill.",
        )
        assert self.skill.skill_context.is_active is False

    def test_init_now_before_start(self):
        """Test the init of Parameters where now is before the registration_start_time."""
        mocked_start_time_str = "01 01 2020  00:03"

        mocked_start_time = self._time(mocked_start_time_str)
        mocked_now_time = self._time("01 01 2020  00:02")

        datetime_mock = Mock(wraps=datetime.datetime)
        datetime_mock.now.return_value = mocked_now_time

        self.kwargs["registration_start_time"] = mocked_start_time_str

        with patch("datetime.datetime", new=datetime_mock):
            with patch.object(self.parameters.context.logger, "log") as mocked_logger:
                par = Parameters(**self.kwargs)
        mocked_logger.assert_any_call(
            logging.INFO,
            f"TAC registation start time: {mocked_start_time}, and registration end time: {par.registration_end_time}, and start time: "
            f"{par.start_time}, "
            f"and end time: {par.end_time}",
        )

    def test_simple_properties(self):
        """Test the properties of Parameters class."""
        # phase
        assert self.parameters.ledger_id == "some_ledger_id"

        self.parameters._contract_address = None
        with pytest.raises(AEAEnforceError, match="No contract address provided."):
            assert self.parameters.contract_address

        self.parameters.contract_address = "some_contract_address"
        assert self.parameters.contract_address == "some_contract_address"

        with pytest.raises(AEAEnforceError, match="Contract address already provided."):
            self.parameters.contract_address = "some_contract_address"

        assert self.parameters.contract_id == self.parameters._contract_id

        assert self.parameters.is_contract_deployed is True
        self.parameters._contract_address = None
        assert self.parameters.is_contract_deployed is False

        assert self.parameters.registration_end_time == datetime.datetime.strptime(
            "01 01 2020  00:02", "%d %m %Y %H:%M"
        )

        assert self.parameters.inactivity_timeout == 30

        assert self.parameters.agent_location == {
            "location": Location(latitude=51.5194, longitude=0.1270)
        }
        assert self.parameters.set_service_data == {"key": "tac", "value": "v1"}
        assert self.parameters.remove_service_data == {"key": "tac"}
        assert self.parameters.simple_service_data == {"tac": "v1"}

    def test_init_inconsistent(self):
        """Test the __init__ of the Parameters class where _check_consistency raises an exception."""
        self.kwargs["contract_address"] = "some_contract_address"
        error_message = "If the contract address is set, then good ids and currency id must be provided and consistent."

        with pytest.raises(ValueError, match=error_message):
            assert Parameters(**self.kwargs,)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains a class representing the game parameters."""

import datetime
from typing import Dict, List, Optional, Set

from aea.exceptions import AEAEnforceError, enforce
from aea.helpers.search.models import Location
from aea.skills.base import Model

from packages.fetchai.contracts.erc1155.contract import PUBLIC_ID as CONTRACT_ID
from packages.fetchai.skills.tac_control.helpers import (
    generate_currency_id_to_name,
    generate_good_id_to_name,
)


DEFAULT_MIN_NB_AGENTS = 2
DEFAULT_MONEY_ENDOWMENT = 200
DEFAULT_NB_GOODS = 9  # ERC1155 vyper contract only accepts 10 tokens per mint/create
DEFAULT_NB_CURRENCIES = 1
DEFAULT_TX_FEE = 1
DEFAULT_BASE_GOOD_ENDOWMENT = 2
DEFAULT_LOWER_BOUND_FACTOR = 1
DEFAULT_UPPER_BOUND_FACTOR = 1
DEFAULT_REGISTRATION_START_TIME = "01 01 2020  00:01"
DEFAULT_REGISTRATION_TIMEOUT = 60
DEFAULT_ITEM_SETUP_TIMEOUT = 60
DEFAULT_COMPETITION_TIMEOUT = 300
DEFAULT_INACTIVITY_TIMEOUT = 30
DEFAULT_LOCATION = {"longitude": 0.1270, "latitude": 51.5194}
DEFAULT_SERVICE_DATA = {"key": "tac", "value": "v1"}


class Parameters(Model):
    """This class contains the parameters of the game."""

    def __init__(self, **kwargs):
        """Instantiate the parameter class."""
        ledger_id = kwargs.pop("ledger_id", None)
        self._contract_address = kwargs.pop(
            "contract_address", None
        )  # type: Optional[str]
        self._good_ids = kwargs.pop("good_ids", [])  # type: List[int]
        self._currency_ids = kwargs.pop("currency_ids", [])  # type: List[int]
        self._min_nb_agents = kwargs.pop(
            "min_nb_agents", DEFAULT_MIN_NB_AGENTS
        )  # type: int
        self._money_endowment = kwargs.pop(
            "money_endowment", DEFAULT_MONEY_ENDOWMENT
        )  # type: int
        self._nb_goods = kwargs.pop("nb_goods", DEFAULT_NB_GOODS)  # type: int
        self._nb_currencies = kwargs.pop(
            "nb_currencies", DEFAULT_NB_CURRENCIES
        )  # type: int
        self._tx_fee = kwargs.pop("tx_fee", DEFAULT_TX_FEE)  # type: int
        self._base_good_endowment = kwargs.pop(
            "base_good_endowment", DEFAULT_BASE_GOOD_ENDOWMENT
        )  # type: int
        self._lower_bound_factor = kwargs.pop(
            "lower_bound_factor", DEFAULT_LOWER_BOUND_FACTOR
        )  # type: int
        self._upper_bound_factor = kwargs.pop(
            "upper_bound_factor", DEFAULT_UPPER_BOUND_FACTOR
        )  # type: int
        registration_start_time = kwargs.pop(
            "registration_start_time", DEFAULT_REGISTRATION_START_TIME
        )  # type: str
        self._registration_start_time = datetime.datetime.strptime(
            registration_start_time, "%d %m %Y %H:%M"
        )  # type: datetime.datetime
        self._registration_timeout = kwargs.pop(
            "registration_timeout", DEFAULT_REGISTRATION_TIMEOUT
        )  # type: int
        self._item_setup_timeout = kwargs.pop(
            "item_setup_timeout", DEFAULT_ITEM_SETUP_TIMEOUT
        )  # type: int
        self._competition_timeout = kwargs.pop(
            "competition_timeout", DEFAULT_COMPETITION_TIMEOUT
        )  # type: int
        self._inactivity_timeout = kwargs.pop(
            "inactivity_timeout", DEFAULT_INACTIVITY_TIMEOUT
        )  # type: int
        self._whitelist = set(kwargs.pop("whitelist", []))  # type: Set[str]
        self._location = kwargs.pop("location", DEFAULT_LOCATION)
        self._service_data = kwargs.pop("service_data", DEFAULT_SERVICE_DATA)
        enforce(
            len(self._service_data) == 2
            and "key" in self._service_data
            and "value" in self._service_data,
            "service_data must contain keys `key` and `value`",
        )
        self._version_id = self._service_data["value"]  # type: str

        self._agent_location = {
            "location": Location(
                latitude=self._location["latitude"],
                longitude=self._location["longitude"],
            )
        }
        self._set_service_data = self._service_data
        self._remove_service_data = {"key": self._service_data["key"]}
        self._simple_service_data = {
            self._service_data["key"]: self._service_data["value"]
        }

        super().__init__(**kwargs)
        self._ledger_id = (
            ledger_id if ledger_id is not None else self.context.default_ledger_id
        )
        self._contract_id = str(CONTRACT_ID)
        self._currency_id_to_name = generate_currency_id_to_name(
            self.nb_currencies, self.currency_ids
        )
        self._good_id_to_name = generate_good_id_to_name(
            self.nb_goods, self.good_ids, starting_index=1
        )
        self._registration_end_time = (
            self._registration_start_time
            + datetime.timedelta(seconds=self._registration_timeout)
        )
        self._start_time = self._registration_end_time + datetime.timedelta(
            seconds=self._item_setup_timeout
        )
        self._end_time = self._start_time + datetime.timedelta(
            seconds=self._competition_timeout
        )
        now = datetime.datetime.now()
        if now > self.registration_start_time:
            self.context.logger.warning(
                "TAC registration start time {} is in the past! Deregistering skill.".format(
                    self.registration_start_time
                )
            )
            self.context.is_active = False
        else:
            self.context.logger.info(
                "TAC registation start time: {}, and registration end time: {}, and start time: {}, and end time: {}".format(
                    self.registration_start_time,
                    self.registration_end_time,
                    self.start_time,
                    self.end_time,
                )
            )
        self._check_consistency()

    @property
    def ledger_id(self) -> str:
        """Get the ledger identifier."""
        return self._ledger_id

    @property
    def contract_address(self) -> str:
        """The contract address of an already deployed smart-contract."""
        if self._contract_address is None:
            raise AEAEnforceError("No contract address provided.")
        return self._contract_address

    @contract_address.setter
    def contract_address(self, contract_address: str) -> None:
        """Set contract address of an already deployed smart-contract."""
        if self._contract_address is not None:
            raise AEAEnforceError("Contract address already provided.")
        self._contract_address = contract_address

    @property
    def contract_id(self) -> str:
        """Get the contract id."""
        return self._contract_id

    @property
    def is_contract_deployed(self) -> bool:
        """Check if there is a deployed instance of the contract."""
        return self._contract_address is not None

    @property
    def good_ids(self) -> List[int]:
        """The item ids of an already deployed smart-contract."""
        return self._good_ids

    @property
    def currency_ids(self) -> List[int]:
        """The currency ids of an already deployed smart-contract."""
        return self._currency_ids

    @property
    def min_nb_agents(self) -> int:
        """Minimum number of agents required for a TAC instance."""
        return self._min_nb_agents

    @property
    def money_endowment(self) -> int:
        """Money endowment per agent for a TAC instance."""
        return self._money_endowment

    @property
    def nb_goods(self) -> int:
        """Good number for a TAC instance."""
        return self._nb_goods

    @property
    def nb_currencies(self) -> int:
        """Currency number for a TAC instance."""
        return self._nb_currencies

    @property
    def currency_id_to_name(self) -> Dict[str, str]:
        """Mapping of currency ids to names"""
        return self._currency_id_to_name

    @property
    def good_id_to_name(self) -> Dict[str, str]:
        """Mapping of good ids to names."""
        return self._good_id_to_name

    @property
    def tx_fee(self) -> int:
        """Transaction fee for a TAC instance."""
        return self._tx_fee

    @property
    def base_good_endowment(self) -> int:
        """Minimum endowment of each agent for each good."""
        return self._base_good_endowment

    @property
    def lower_bound_factor(self) -> int:
        """Lower bound of a uniform distribution."""
        return self._lower_bound_factor

    @property
    def upper_bound_factor(self) -> int:
        """Upper bound of a uniform distribution."""
        return self._upper_bound_factor

    @property
    def registration_start_time(self) -> datetime.datetime:
        """TAC registration start time."""
        return self._registration_start_time

    @property
    def registration_end_time(self) -> datetime.datetime:
        """TAC registration end time."""
        return self._registration_end_time

    @property
    def start_time(self) -> datetime.datetime:
        """TAC start time."""
        return self._start_time

    @property
    def end_time(self) -> datetime.datetime:
        """TAC end time."""
        return self._end_time

    @property
    def inactivity_timeout(self):
        """Timeout of agent inactivity from controller perspective (no received transactions)."""
        return self._inactivity_timeout

    @property
    def whitelist(self) -> Set[str]:
        """Whitelist of agent addresses allowed into the TAC instance."""
        return self._whitelist

    @property
    def version_id(self) -> str:
        """Version id."""
        return self._version_id

    @property
    def agent_location(self) -> Dict[str, Location]:
        """Get the agent location."""
        return self._agent_location

    @property
    def set_service_data(self) -> Dict[str, str]:
        """Get the set service data."""
        return self._set_service_data

    @property
    def remove_service_data(self) -> Dict[str, str]:
        """Get the remove service data."""
        return self._remove_service_data

    @property
    def simple_service_data(self) -> Dict[str, str]:
        """Get the simple service data."""
        return self._simple_service_data

    def _check_consistency(self) -> None:
        """Check the parameters are consistent."""
        if self._contract_address is not None and (
            (self._good_ids is not None and len(self._good_ids) != self._nb_goods)
            or (
                self._currency_ids is not None
                and len(self._currency_ids) != self._nb_currencies
            )
        ):
            raise ValueError(
                "If the contract address is set, then good ids and currency id must be provided and consistent."
            )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains tests for CLI Registry login methods."""

from unittest import TestCase, mock

from aea.cli.registry.login import registry_login, registry_reset_password


@mock.patch("aea.cli.registry.login.request_api", return_value={"key": "key"})
class RegistryLoginTestCase(TestCase):
    """Test case for registry_login method."""

    def test_registry_login_positive(self, request_api_mock):
        """Test for registry_login method positive result."""
        result = registry_login("username", "password")
        expected_result = "key"
        self.assertEqual(result, expected_result)
        request_api_mock.assert_called_once()


@mock.patch(
    "aea.cli.registry.login.request_api", return_value={"message": "Email was sent."}
)
class RegistryResetPasswordTestCase(TestCase):
    """Test case for registry_reset_password method."""

    def test_registry_reset_password_positive(self, request_api_mock):
        """Test for registry_reset_password method positive result."""
        registry_reset_password("email@example.com")
        request_api_mock.assert_called_once()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the models of the tac participation skill."""

import logging
from pathlib import Path
from typing import cast
from unittest.mock import patch

import pytest

from aea.exceptions import AEAEnforceError
from aea.helpers.search.models import Constraint, ConstraintType, Location, Query
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.protocols.state_update.message import StateUpdateMessage
from packages.fetchai.protocols.tac.message import TacMessage
from packages.fetchai.skills.tac_participation.dialogues import (
    StateUpdateDialogues,
    TacDialogues,
)
from packages.fetchai.skills.tac_participation.game import Configuration, Game, Phase

from tests.conftest import ROOT_DIR


class TestConfiguration:
    """Test Configuration class of tac participation."""

    @classmethod
    def setup(cls):
        """Setup the test class."""
        cls.version_id = "some_version_id"
        cls.fee_by_currency_id = {"1": 1}
        cls.agent_addr_to_name = {
            "agent_address_1": "agent_name_1",
            "agent_address_2": "agent_name_2",
        }
        cls.good_id_to_name = {"3": "good_1", "4": "good_2"}
        cls.controller_addr = "some_controller_address"

        cls.configuration = Configuration(
            cls.version_id,
            cls.fee_by_currency_id,
            cls.agent_addr_to_name,
            cls.good_id_to_name,
            cls.controller_addr,
        )

    def test_simple_properties(self):
        """Test the properties of Game class."""
        assert self.configuration.version_id == self.version_id

        assert self.configuration.nb_agents == len(self.agent_addr_to_name)

        assert self.configuration.nb_goods == len(self.good_id_to_name)

        assert self.configuration.tx_fee == 1

        self.configuration._fee_by_currency_id = {"1": 1, "2": 2}
        with pytest.raises(AEAEnforceError, match="More than one currency id present!"):
            assert self.configuration.tx_fee
        self.configuration._fee_by_currency_id = self.fee_by_currency_id

        assert self.configuration.fee_by_currency_id == self.fee_by_currency_id

        assert self.configuration.agent_addr_to_name == self.agent_addr_to_name

        assert self.configuration.good_id_to_name == self.good_id_to_name

        assert self.configuration.agent_addresses == list(
            self.agent_addr_to_name.keys()
        )

        assert self.configuration.agent_names == list(self.agent_addr_to_name.values())

        assert self.configuration.good_ids == list(self.good_id_to_name.keys())

        assert self.configuration.good_names == list(self.good_id_to_name.values())

        assert self.configuration.controller_addr == self.controller_addr

    def test_check_consistency_succeeds(self):
        """Test the _check_consistency of Configuration class which succeeds."""
        self.configuration._check_consistency()

    def test_check_consistency_fails_i(self):
        """Test the _check_consistency of Configuration class which fails on version being None."""
        self.configuration._version_id = None
        with pytest.raises(AEAEnforceError, match="A version id must be set."):
            assert self.configuration._check_consistency()

    def test_check_consistency_fails_ii(self):
        """Test the _check_consistency of Configuration class which fails because _fee_by_currency_id has more than one currencies."""
        self.configuration._fee_by_currency_id = {"1": 1, "2": 2}
        with pytest.raises(AEAEnforceError, match="Tx fee must be non-negative."):
            assert self.configuration._check_consistency()

    def test_check_consistency_fails_iii(self):
        """Test the _check_consistency of Configuration class which fails because tx_fee < 0."""
        self.configuration._fee_by_currency_id = {"1": -5}
        with pytest.raises(AEAEnforceError, match="Tx fee must be non-negative."):
            assert self.configuration._check_consistency()

    def test_check_consistency_fails_iv(self):
        """Test the _check_consistency of Configuration class which fails because number of agents is less than 2."""
        incorrect_agent_addr_to_name = {"agent_address_1": "agent_name_1"}
        with pytest.raises(AEAEnforceError, match="Must have at least two agents."):
            Configuration(
                self.version_id,
                self.fee_by_currency_id,
                incorrect_agent_addr_to_name,
                self.good_id_to_name,
                self.controller_addr,
            )

    def test_check_consistency_fails_v(self):
        """Test the _check_consistency of Configuration class which fails because number of goods is less than 2."""
        incorrect_good_id_to_name = {"3": "good_1"}
        with pytest.raises(AEAEnforceError, match="Must have at least two goods."):
            Configuration(
                self.version_id,
                self.fee_by_currency_id,
                self.agent_addr_to_name,
                incorrect_good_id_to_name,
                self.controller_addr,
            )

    def test_check_consistency_fails_vi(self):
        """Test the _check_consistency of Configuration class which fails because number of goods is less than 2."""
        self.configuration._agent_addr_to_name = {"agent_address_1": "agent_name_1"}
        with pytest.raises(
            AEAEnforceError, match="There must be one address for each agent."
        ):
            self.configuration._check_consistency()

    def test_check_consistency_fails_vii(self):
        """Test the _check_consistency of Configuration class which fails because number of goods is less than 2."""
        incorrect_agent_addr_to_name = {
            "agent_address_1": "agent_name_1",
            "agent_address_2": "agent_name_1",
        }
        with pytest.raises(AEAEnforceError, match="Agents' names must be unique."):
            Configuration(
                self.version_id,
                self.fee_by_currency_id,
                incorrect_agent_addr_to_name,
                self.good_id_to_name,
                self.controller_addr,
            )

    def test_check_consistency_fails_viii(self):
        """Test the _check_consistency of Configuration class which fails because number of goods is less than 2."""
        self.configuration._good_id_to_name = {
            "3": "good_1",
            "4": "good_2",
            "5": "good_3",
        }
        with pytest.raises(
            AEAEnforceError, match="There must be one id for each good."
        ):
            self.configuration._check_consistency()

    def test_check_consistency_fails_ix(self):
        """Test the _check_consistency of Configuration class which fails because number of goods is less than 2."""
        incorrect_good_id_to_name = {"3": "good_1", "4": "good_1"}
        with pytest.raises(AEAEnforceError, match="Goods' names must be unique."):
            Configuration(
                self.version_id,
                self.fee_by_currency_id,
                self.agent_addr_to_name,
                incorrect_good_id_to_name,
                self.controller_addr,
            )


class TestGame(BaseSkillTestCase):
    """Test Game class of tac participation."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_participation")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.expected_version_id = "some_version_id"
        cls.expected_controller_addr = "some_controller_address"
        cls.search_query = {
            "search_key": "tac",
            "search_value": "v1",
            "constraint_type": "==",
        }
        cls.location = {"longitude": 0.1270, "latitude": 51.5194}
        cls.search_radius = 5.0
        cls.ledger_id = "some_ledger_id"
        cls.is_using_contract = False

        cls.game = Game(
            expected_version_id=cls.expected_version_id,
            expected_controller_addr=cls.expected_controller_addr,
            search_query=cls.search_query,
            location=cls.location,
            search_radius=cls.search_radius,
            ledger_id=cls.ledger_id,
            is_using_contract=cls.is_using_contract,
            name="game",
            skill_context=cls._skill.skill_context,
        )

    def test_simple_properties(self):
        """Test the properties of Game class."""
        assert self.game.ledger_id == self.ledger_id

        assert self.game.is_using_contract == self.is_using_contract

        assert self.game.expected_version_id == self.expected_version_id

        assert self.game.phase == Phase.PRE_GAME

        with pytest.raises(AEAEnforceError, match="Contract address not set!"):
            assert self.game.contract_address
        self.game.contract_address = "some_contract_address"
        assert self.game.contract_address == "some_contract_address"
        with pytest.raises(AEAEnforceError, match="Contract address already set!"):
            self.game.contract_address = "some_other_contract_address"

        with pytest.raises(AEAEnforceError, match="TacDialogue not set!"):
            assert self.game.tac_dialogue
        _, tac_dialogue = cast(
            TacDialogues, self.skill.skill_context.tac_dialogues
        ).create(
            counterparty="some_address",
            performative=TacMessage.Performative.REGISTER,
            agent_name="some_agent_name",
        )
        self.game.tac_dialogue = tac_dialogue
        assert self.game.tac_dialogue == tac_dialogue
        with pytest.raises(AEAEnforceError, match="TacDialogue already set!"):
            self.game.tac_dialogue = tac_dialogue

        with pytest.raises(AEAEnforceError, match="StateUpdateDialogue not set!"):
            assert self.game.state_update_dialogue
        _, state_update_dialogue = cast(
            StateUpdateDialogues, self.skill.skill_context.state_update_dialogues
        ).create(
            counterparty="some_address",
            performative=StateUpdateMessage.Performative.INITIALIZE,
            exchange_params_by_currency_id={"some_currency_id": 1.0},
            utility_params_by_good_id={"some_good_id": 2.0},
            amount_by_currency_id={"some_currency_id": 10},
            quantities_by_good_id={"some_good_id": 5},
        )
        self.game.state_update_dialogue = state_update_dialogue
        assert self.game.state_update_dialogue == state_update_dialogue
        with pytest.raises(AEAEnforceError, match="StateUpdateDialogue already set!"):
            self.game.state_update_dialogue = state_update_dialogue

        assert self.game.expected_controller_addr == self.expected_controller_addr
        self.game._expected_controller_addr = None
        with pytest.raises(
            AEAEnforceError, match="Expected controller address not assigned!"
        ):
            assert self.game.expected_controller_addr

        with pytest.raises(AEAEnforceError, match="Game configuration not assigned!"):
            assert self.game.conf
        configuration = Configuration(
            "some_version_id",
            {"1": 1},
            {"agent_address_1": "agent_name_1", "agent_address_2": "agent_name_2"},
            {"3": "good_1", "4": "good_2"},
            "some_controller_address",
        )
        self.game._conf = configuration
        assert self.game.conf == configuration

    def test_init_succeeds(self):
        """Test the init method of the Game class which succeeds."""
        fee_by_currency_id = {"1": 1}
        agent_addr_to_name = {
            "some_address_1": "some_name_1",
            "some_address_2": "some_name_2",
        }
        good_id_to_name = {"2": "good_2", "3": "good_3"}
        tac_message = cast(
            TacMessage,
            self.build_incoming_message(
                message_type=TacMessage,
                performative=TacMessage.Performative.GAME_DATA,
                amount_by_currency_id={"1": 10},
                exchange_params_by_currency_id={"1": 1.0},
                quantities_by_good_id={"2": 10, "3": 11},
                utility_params_by_good_id={"2": 1.2, "3": 1.1},
                fee_by_currency_id=fee_by_currency_id,
                agent_addr_to_name=agent_addr_to_name,
                currency_id_to_name={"1": "FETCH"},
                good_id_to_name=good_id_to_name,
                version_id=self.expected_version_id,
            ),
        )

        self.game.init(tac_message, self.expected_controller_addr)

        assert self.game.conf is not None
        assert self.game.conf.version_id == self.expected_version_id
        assert self.game.conf.fee_by_currency_id == fee_by_currency_id
        assert self.game.conf.agent_addr_to_name == agent_addr_to_name
        assert self.game.conf.good_id_to_name == good_id_to_name
        assert self.game.conf.controller_addr == self.expected_controller_addr

    def test_init_fails_i(self):
        """Test the init method of the Game class which fails because performative is NOT GAME_DATA."""
        tac_message = cast(
            TacMessage,
            self.build_incoming_message(
                message_type=TacMessage,
                performative=TacMessage.Performative.REGISTER,
                agent_name="some_agent_name",
            ),
        )

        with pytest.raises(
            AEAEnforceError, match="Wrong TacMessage for initialization of TAC game."
        ):
            self.game.init(tac_message, self.expected_controller_addr)

    def test_init_fails_ii(self):
        """Test the init method of the Game class which fails because controller address is incorrect."""
        fee_by_currency_id = {"1": 1}
        agent_addr_to_name = {
            "some_address_1": "some_name_1",
            "some_address_2": "some_name_2",
        }
        good_id_to_name = {"2": "good_2", "3": "good_3"}
        incorrect_controller_addr = "some_other_controller_address"
        tac_message = cast(
            TacMessage,
            self.build_incoming_message(
                message_type=TacMessage,
                performative=TacMessage.Performative.GAME_DATA,
                amount_by_currency_id={"1": 10},
                exchange_params_by_currency_id={"1": 1.0},
                quantities_by_good_id={"2": 10, "3": 11},
                utility_params_by_good_id={"2": 1.2, "3": 1.1},
                fee_by_currency_id=fee_by_currency_id,
                agent_addr_to_name=agent_addr_to_name,
                currency_id_to_name={"1": "FETCH"},
                good_id_to_name=good_id_to_name,
                version_id=self.expected_version_id,
            ),
        )

        with pytest.raises(
            AEAEnforceError, match="TacMessage from unexpected controller."
        ):
            self.game.init(tac_message, incorrect_controller_addr)

    def test_init_fails_iii(self):
        """Test the init method of the Game class which fails because version id is incorrect."""
        fee_by_currency_id = {"1": 1}
        agent_addr_to_name = {
            "some_address_1": "some_name_1",
            "some_address_2": "some_name_2",
        }
        good_id_to_name = {"2": "good_2", "3": "good_3"}
        incorrect_version_id = "some_other_version_id"
        tac_message = cast(
            TacMessage,
            self.build_incoming_message(
                message_type=TacMessage,
                performative=TacMessage.Performative.GAME_DATA,
                amount_by_currency_id={"1": 10},
                exchange_params_by_currency_id={"1": 1.0},
                quantities_by_good_id={"2": 10, "3": 11},
                utility_params_by_good_id={"2": 1.2, "3": 1.1},
                fee_by_currency_id=fee_by_currency_id,
                agent_addr_to_name=agent_addr_to_name,
                currency_id_to_name={"1": "FETCH"},
                good_id_to_name=good_id_to_name,
                version_id=incorrect_version_id,
            ),
        )

        with pytest.raises(AEAEnforceError, match="TacMessage for unexpected game."):
            self.game.init(tac_message, self.expected_controller_addr)

    def test_update_expected_controller_addr(self):
        """Test the update_expected_controller_addr method of the Game class."""
        new_contract_address = "some_different_controller_address"
        with patch.object(self.skill.skill_context.logger, "log") as mock_logger:
            self.game.update_expected_controller_addr(new_contract_address)

        mock_logger.assert_any_call(
            logging.WARNING,
            "TAKE CARE! Circumventing controller identity check! For added security provide the expected controller key as an argument to the Game instance "
            "and check against it.",
        )
        assert self.game.expected_controller_addr == new_contract_address

    def test_update_game_phase(self):
        """Test the update_game_phase method of the Game class."""
        assert self.game.phase == Phase.PRE_GAME
        self.game.update_game_phase(Phase.GAME)
        assert self.game.phase == Phase.GAME

    def test_get_game_query(self):
        """Test the get_game_query method of the Game class."""
        expected_location = Location(
            latitude=self.location["latitude"], longitude=self.location["longitude"]
        )
        expected_query = Query(
            [
                Constraint(
                    "location",
                    ConstraintType("distance", (expected_location, self.search_radius)),
                ),
                Constraint(
                    self.search_query["search_key"],
                    ConstraintType(
                        self.search_query["constraint_type"],
                        self.search_query["search_value"],
                    ),
                ),
            ],
        )

        actual_query = self.game.get_game_query()
        assert actual_query == expected_query



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains tests for the register protocol."""
from typing import Type
from unittest.mock import patch

import pytest

from aea.common import Address
from aea.protocols.base import Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue
from aea.protocols.dialogue.base import DialogueLabel

from packages.fetchai.protocols.register.dialogues import (
    RegisterDialogue as BaseRegisterDialogue,
)
from packages.fetchai.protocols.register.dialogues import (
    RegisterDialogues as BaseRegisterDialogues,
)
from packages.fetchai.protocols.register.message import RegisterMessage


class TestRegisterMessage:
    """Test the register message module."""

    @classmethod
    def setup_class(cls):
        """Setup class for test case."""
        cls.info = {"a": "b", "c": "d"}

    def test_register(self):
        """Test for an error for a register message."""
        tx_msg = RegisterMessage(
            performative=RegisterMessage.Performative.REGISTER, info=self.info
        )
        assert tx_msg._is_consistent()
        encoded_tx_msg = tx_msg.encode()
        decoded_tx_msg = tx_msg.serializer.decode(encoded_tx_msg)
        assert tx_msg == decoded_tx_msg

    def test_success(self):
        """Test for an error for a success message."""
        tx_msg = RegisterMessage(
            performative=RegisterMessage.Performative.SUCCESS,
            info=self.info,
            target=1,
            message_id=2,
        )
        assert tx_msg._is_consistent()
        encoded_tx_msg = tx_msg.encode()
        decoded_tx_msg = tx_msg.serializer.decode(encoded_tx_msg)
        assert tx_msg == decoded_tx_msg

    def test_error(self):
        """Test for an error for a register error message."""
        some_error_code = 1
        some_error_msg = "Some error message"
        tx_msg = RegisterMessage(
            performative=RegisterMessage.Performative.ERROR,
            error_code=some_error_code,
            error_msg=some_error_msg,
            info=self.info,
        )
        assert tx_msg._is_consistent()
        encoded_tx_msg = tx_msg.encode()
        decoded_tx_msg = tx_msg.serializer.decode(encoded_tx_msg)
        assert tx_msg == decoded_tx_msg


def test_consistency_check_negative():
    """Test the consistency check, negative case."""
    tx_msg = RegisterMessage(performative=RegisterMessage.Performative.REGISTER,)
    assert not tx_msg._is_consistent()


def test_serialization_negative():
    """Test serialization when performative is not recognized."""
    tx_msg = RegisterMessage(
        performative=RegisterMessage.Performative.REGISTER, info={},
    )

    with patch.object(RegisterMessage.Performative, "__eq__", return_value=False):
        with pytest.raises(
            ValueError, match=f"Performative not valid: {tx_msg.performative}"
        ):
            tx_msg.serializer.encode(tx_msg)

    encoded_tx_bytes = tx_msg.serializer.encode(tx_msg)
    with patch.object(RegisterMessage.Performative, "__eq__", return_value=False):
        with pytest.raises(
            ValueError, match=f"Performative not valid: {tx_msg.performative}"
        ):
            tx_msg.serializer.decode(encoded_tx_bytes)


def test_dialogues():
    """Test instantiation of dialogues."""
    register_dialogues = RegisterDialogues("agent_addr")
    msg, dialogue = register_dialogues.create(
        counterparty="abc", performative=RegisterMessage.Performative.REGISTER, info={}
    )
    assert dialogue is not None


class RegisterDialogue(BaseRegisterDialogue):
    """The dialogue class maintains state of a dialogue and manages it."""

    def __init__(
        self,
        dialogue_label: DialogueLabel,
        self_address: Address,
        role: BaseDialogue.Role,
        message_class: Type[RegisterMessage],
    ) -> None:
        """
        Initialize a dialogue.

        :param dialogue_label: the identifier of the dialogue
        :param self_address: the address of the entity for whom this dialogue is maintained
        :param role: the role of the agent this dialogue is maintained for

        :return: None
        """
        BaseRegisterDialogue.__init__(
            self,
            dialogue_label=dialogue_label,
            self_address=self_address,
            role=role,
            message_class=message_class,
        )


class RegisterDialogues(BaseRegisterDialogues):
    """The dialogues class keeps track of all dialogues."""

    def __init__(self, self_address: Address) -> None:
        """
        Initialize dialogues.

        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return RegisterDialogue.Role.AGENT

        BaseRegisterDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
            dialogue_class=RegisterDialogue,
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the handler classes of the tac control skill."""

import logging
from pathlib import Path
from typing import cast
from unittest.mock import patch

from aea.crypto.ledger_apis import LedgerApis
from aea.helpers.transaction.base import (
    RawTransaction,
    SignedTransaction,
    State,
    Terms,
    TransactionDigest,
    TransactionReceipt,
)
from aea.protocols.dialogue.base import DialogueMessage
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.protocols.contract_api.custom_types import Kwargs
from packages.fetchai.protocols.contract_api.message import ContractApiMessage
from packages.fetchai.protocols.ledger_api.message import LedgerApiMessage
from packages.fetchai.protocols.signing.message import SigningMessage
from packages.fetchai.skills.tac_control_contract.dialogues import (
    ContractApiDialogue,
    ContractApiDialogues,
    LedgerApiDialogue,
    LedgerApiDialogues,
    SigningDialogue,
    SigningDialogues,
)
from packages.fetchai.skills.tac_control_contract.game import Game, Phase
from packages.fetchai.skills.tac_control_contract.handlers import (
    ContractApiHandler,
    LEDGER_API_ADDRESS,
    LedgerApiHandler,
    SigningHandler,
)
from packages.fetchai.skills.tac_control_contract.parameters import Parameters

from tests.conftest import ROOT_DIR


class TestContractApiHandler(BaseSkillTestCase):
    """Test contract_api handler of tac control contract."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "fetchai", "skills", "tac_control_contract"
    )

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.contract_api_handler = cast(
            ContractApiHandler, cls._skill.skill_context.handlers.contract_api
        )
        cls.contract_api_dialogues = cast(
            ContractApiDialogues, cls._skill.skill_context.contract_api_dialogues
        )
        cls.signing_dialogues = cast(
            SigningDialogues, cls._skill.skill_context.signing_dialogues
        )
        cls.logger = cls.contract_api_handler.context.logger

        cls.ledger_id = "some_ledger_id"
        cls.contract_id = "some_contract_id"
        cls.callable = "some_callable"
        cls.kwargs = Kwargs({"some_key": "some_value"})

        cls.list_of_contract_api_messages = (
            DialogueMessage(
                ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
                {
                    "ledger_id": cls.ledger_id,
                    "contract_id": cls.contract_id,
                    "callable": cls.callable,
                    "kwargs": cls.kwargs,
                },
            ),
        )

    def test_setup(self):
        """Test the setup method of the contract_api handler."""
        assert self.contract_api_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the signing handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=ContractApiMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=ContractApiMessage.Performative.STATE,
            state=State("some_ledger_id", b"some_body"),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.contract_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid contract_api message={incoming_message}, unidentified dialogue.",
        )

    def test_handle_raw_transaction(self,):
        """Test the _handle_signed_transaction method of the signing handler."""
        # setup
        contract_api_dialogue = cast(
            ContractApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.contract_api_dialogues,
                messages=self.list_of_contract_api_messages[:1],
            ),
        )
        contract_api_dialogue.terms = Terms(
            "some_ledger_id",
            self.skill.skill_context.agent_address,
            "counterprty",
            {"currency_id": 50},
            {"good_id": -10},
            "some_nonce",
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=contract_api_dialogue,
            performative=ContractApiMessage.Performative.RAW_TRANSACTION,
            raw_transaction=ContractApiMessage.RawTransaction(
                "some_ledger_id", "some_body"
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.contract_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO, f"received raw transaction={incoming_message}"
        )

        self.assert_quantity_in_decision_making_queue(1)
        message = self.get_message_from_decision_maker_inbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=SigningMessage,
            performative=SigningMessage.Performative.SIGN_TRANSACTION,
            to=self.skill.skill_context.decision_maker_address,
            sender=str(self.skill.skill_context.skill_id),
            terms=contract_api_dialogue.terms,
        )
        assert has_attributes, error_str

        assert (
            cast(
                SigningDialogue, self.signing_dialogues.get_dialogue(message)
            ).associated_contract_api_dialogue
            == contract_api_dialogue
        )

        mock_logger.assert_any_call(
            logging.INFO,
            "proposing the transaction to the decision maker. Waiting for confirmation ...",
        )

    def test_handle_error(self):
        """Test the _handle_error method of the signing handler."""
        # setup
        contract_api_dialogue = self.prepare_skill_dialogue(
            dialogues=self.contract_api_dialogues,
            messages=self.list_of_contract_api_messages[:1],
        )
        incoming_message = cast(
            ContractApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=contract_api_dialogue,
                performative=ContractApiMessage.Performative.ERROR,
                data=b"some_data",
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.contract_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received contract_api error message={incoming_message} in dialogue={contract_api_dialogue}.",
        )

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the signing handler."""
        # setup
        invalid_performative = ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION
        incoming_message = self.build_incoming_message(
            message_type=ContractApiMessage,
            dialogue_reference=("1", ""),
            performative=invalid_performative,
            ledger_id=self.ledger_id,
            contract_id=self.contract_id,
            callable=self.callable,
            kwargs=self.kwargs,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.contract_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle contract_api message of performative={invalid_performative} in dialogue={self.contract_api_dialogues.get_dialogue(incoming_message)}.",
        )

    def test_teardown(self):
        """Test the teardown method of the contract_api handler."""
        assert self.contract_api_handler.teardown() is None
        self.assert_quantity_in_outbox(0)


class TestSigningHandler(BaseSkillTestCase):
    """Test signing handler of tac control contract."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "fetchai", "skills", "tac_control_contract"
    )

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.signing_handler = cast(
            SigningHandler, cls._skill.skill_context.handlers.signing
        )
        cls.ledger_api_dialogues = cast(
            LedgerApiDialogues, cls._skill.skill_context.ledger_api_dialogues
        )
        cls.signing_dialogues = cast(
            SigningDialogues, cls._skill.skill_context.signing_dialogues
        )
        cls.logger = cls.signing_handler.context.logger

        cls.terms = Terms(
            "some_ledger_id",
            cls._skill.skill_context.agent_address,
            "counterprty",
            {"currency_id": 50},
            {"good_id": -10},
            "some_nonce",
        )
        cls.list_of_signing_messages = (
            DialogueMessage(
                SigningMessage.Performative.SIGN_TRANSACTION,
                {
                    "terms": cls.terms,
                    "raw_transaction": SigningMessage.RawTransaction(
                        "some_ledger_id", "some_body"
                    ),
                },
            ),
        )

    def test_setup(self):
        """Test the setup method of the signing handler."""
        assert self.signing_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the signing handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=SigningMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=SigningMessage.Performative.ERROR,
            error_code=SigningMessage.ErrorCode.UNSUCCESSFUL_MESSAGE_SIGNING,
            to=str(self.skill.skill_context.skill_id),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid signing message={incoming_message}, unidentified dialogue.",
        )

    def test_handle_signed_transaction(self,):
        """Test the _handle_signed_transaction method of the signing handler."""
        # setup
        signing_counterparty = self.skill.skill_context.decision_maker_address
        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_messages[:1],
                counterparty=signing_counterparty,
            ),
        )
        incoming_message = cast(
            SigningMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=signing_dialogue,
                performative=SigningMessage.Performative.SIGNED_TRANSACTION,
                signed_transaction=SigningMessage.SignedTransaction(
                    "some_ledger_id", "some_body"
                ),
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(logging.INFO, "transaction signing was successful.")

        self.assert_quantity_in_outbox(1)
        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=LedgerApiMessage,
            performative=LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,
            to=LEDGER_API_ADDRESS,
            sender=self.skill.skill_context.agent_address,
            signed_transaction=incoming_message.signed_transaction,
        )
        assert has_attributes, error_str

        assert (
            cast(
                LedgerApiDialogue, self.ledger_api_dialogues.get_dialogue(message)
            ).associated_signing_dialogue
            == signing_dialogue
        )

        mock_logger.assert_any_call(logging.INFO, "sending transaction to ledger.")

    def test_handle_error(self):
        """Test the _handle_error method of the signing handler."""
        # setup
        signing_counterparty = self.skill.skill_context.decision_maker_address
        signing_dialogue = self.prepare_skill_dialogue(
            dialogues=self.signing_dialogues,
            messages=self.list_of_signing_messages[:1],
            counterparty=signing_counterparty,
        )
        incoming_message = cast(
            SigningMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=signing_dialogue,
                performative=SigningMessage.Performative.ERROR,
                error_code=SigningMessage.ErrorCode.UNSUCCESSFUL_TRANSACTION_SIGNING,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"transaction signing was not successful. Error_code={incoming_message.error_code} in dialogue={signing_dialogue}",
        )

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the signing handler."""
        # setup
        invalid_performative = SigningMessage.Performative.SIGN_TRANSACTION
        incoming_message = self.build_incoming_message(
            message_type=SigningMessage,
            dialogue_reference=("1", ""),
            performative=invalid_performative,
            terms=self.terms,
            raw_transaction=SigningMessage.RawTransaction(
                "some_ledger_id", "some_body"
            ),
            to=str(self.skill.skill_context.skill_id),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle signing message of performative={invalid_performative} in dialogue={self.signing_dialogues.get_dialogue(incoming_message)}.",
        )

    def test_teardown(self):
        """Test the teardown method of the signing handler."""
        assert self.signing_handler.teardown() is None
        self.assert_quantity_in_outbox(0)


class TestGenericLedgerApiHandler(BaseSkillTestCase):
    """Test ledger_api handler of tac control contract."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "fetchai", "skills", "tac_control_contract"
    )

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.ledger_api_handler = cast(
            LedgerApiHandler, cls._skill.skill_context.handlers.ledger_api
        )
        cls.ledger_api_dialogues = cast(
            LedgerApiDialogues, cls._skill.skill_context.ledger_api_dialogues
        )
        cls.signing_dialogues = cast(
            SigningDialogues, cls._skill.skill_context.signing_dialogues
        )
        cls.contract_api_dialogues = cast(
            ContractApiDialogues, cls._skill.skill_context.contract_api_dialogues
        )
        cls.parameters = cast(Parameters, cls._skill.skill_context.parameters)
        cls.game = cast(Game, cls._skill.skill_context.game)

        cls.logger = cls.ledger_api_handler.context.logger

        cls.ledger_id = "some_ledger_id"
        cls.contract_id = "some_contract_id"
        cls.callable = "some_callable"
        cls.kwargs = Kwargs({"some_key": "some_value"})

        cls.body = "some_body"
        cls.contract_address = "some_contract_address"

        cls.raw_transaction = RawTransaction(cls.ledger_id, cls.body)
        cls.signed_transaction = SignedTransaction(cls.ledger_id, cls.body)
        cls.transaction_digest = TransactionDigest(cls.ledger_id, cls.body)
        cls.receipt = {"contractAddress": cls.contract_address}
        cls.transaction_receipt = TransactionReceipt(
            cls.ledger_id, cls.receipt, "some_transaction"
        )

        cls.terms = Terms(
            cls.ledger_id,
            cls._skill.skill_context.agent_address,
            "counterprty",
            {"currency_id": 50},
            {"good_id": -10},
            "some_nonce",
        )

        cls.list_of_signing_messages = (
            DialogueMessage(
                SigningMessage.Performative.SIGN_TRANSACTION,
                {
                    "terms": cls.terms,
                    "raw_transaction": SigningMessage.RawTransaction(
                        cls.ledger_id, cls.body
                    ),
                },
            ),
        )
        cls.list_of_contract_api_messages = (
            DialogueMessage(
                ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
                {
                    "ledger_id": cls.ledger_id,
                    "contract_id": cls.contract_id,
                    "callable": cls.callable,
                    "kwargs": cls.kwargs,
                },
            ),
        )
        cls.list_of_ledger_api_messages = (
            DialogueMessage(
                LedgerApiMessage.Performative.GET_RAW_TRANSACTION, {"terms": cls.terms}
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.RAW_TRANSACTION,
                {"raw_transaction": cls.raw_transaction},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,
                {"signed_transaction": cls.signed_transaction},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.TRANSACTION_DIGEST,
                {"transaction_digest": cls.transaction_digest},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,
                {"transaction_digest": cls.transaction_digest},
            ),
        )

    def test_setup(self):
        """Test the setup method of the ledger_api handler."""
        assert self.ledger_api_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the ledger_api handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=LedgerApiMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=LedgerApiMessage.Performative.BALANCE,
            ledger_id="some_ledger_id",
            balance=10,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid ledger_api message={incoming_message}, unidentified dialogue.",
        )

    def test_handle_balance(self):
        """Test the _handle_balance method of the ledger_api handler."""
        # setup
        balance = 10
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=(
                    DialogueMessage(
                        LedgerApiMessage.Performative.GET_BALANCE,
                        {"ledger_id": self.ledger_id, "address": "some_address"},
                    ),
                ),
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.BALANCE,
                ledger_id=self.ledger_id,
                balance=balance,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"starting balance on {self.ledger_id} ledger={incoming_message.balance}.",
        )

    def test_handle_transaction_digest(self):
        """Test the _handle_transaction_digest method of the ledger_api handler."""
        # setup
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[:3],
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.TRANSACTION_DIGEST,
                transaction_digest=self.transaction_digest,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"transaction was successfully submitted. Transaction digest={incoming_message.transaction_digest}",
        )

        self.assert_quantity_in_outbox(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=LedgerApiMessage,
            performative=LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,
            to=LEDGER_API_ADDRESS,
            sender=self.skill.skill_context.agent_address,
            transaction_digest=incoming_message.transaction_digest,
        )
        assert has_attributes, error_str

        mock_logger.assert_any_call(logging.INFO, "requesting transaction receipt.")

    def test_handle_transaction_receipt_failed(self):
        """Test the _handle_transaction_receipt method of the ledger_api handler where the transaction is NOT settled."""
        # setup
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[:5],
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_messages[:4],
            ),
        )
        contract_api_dialogue = cast(
            ContractApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_messages[:4],
            ),
        )

        signing_dialogue.associated_contract_api_dialogue = contract_api_dialogue
        ledger_api_dialogue.associated_signing_dialogue = signing_dialogue

        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                transaction_receipt=self.transaction_receipt,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with patch.object(LedgerApis, "is_transaction_settled", return_value=False):
                self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.ERROR,
            f"transaction failed. Transaction receipt={incoming_message.transaction_receipt}",
        )

    def test_handle_transaction_receipt_callable_get_deploy_transaction(self):
        """Test the _handle_transaction_receipt method of the ledger_api handler where contract_api callable is GET_DEPLOY_TRANSACTION."""
        # setup
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[:5],
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_messages[:4],
            ),
        )
        contract_api_dialogue = cast(
            ContractApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_messages[:4],
            ),
        )

        signing_dialogue.associated_contract_api_dialogue = contract_api_dialogue
        ledger_api_dialogue.associated_signing_dialogue = signing_dialogue

        contract_api_dialogue.callable = (
            ContractApiDialogue.Callable.GET_DEPLOY_TRANSACTION
        )

        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                transaction_receipt=self.transaction_receipt,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with patch.object(LedgerApis, "is_transaction_settled", return_value=True):
                self.ledger_api_handler.handle(incoming_message)

        # after
        assert self.parameters.contract_address == self.contract_address
        assert self.game.phase == Phase.CONTRACT_DEPLOYED

        mock_logger.assert_any_call(logging.INFO, "contract deployed.")

    def test_handle_transaction_receipt_callable_get_create_batch_transaction(self):
        """Test the _handle_transaction_receipt method of the ledger_api handler where contract_api callable is GET_CREATE_BATCH_TRANSACTION."""
        # setup
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[:5],
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_messages[:4],
            ),
        )
        contract_api_dialogue = cast(
            ContractApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_messages[:4],
            ),
        )

        signing_dialogue.associated_contract_api_dialogue = contract_api_dialogue
        ledger_api_dialogue.associated_signing_dialogue = signing_dialogue

        contract_api_dialogue.callable = (
            ContractApiDialogue.Callable.GET_CREATE_BATCH_TRANSACTION
        )

        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                transaction_receipt=self.transaction_receipt,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with patch.object(LedgerApis, "is_transaction_settled", return_value=True):
                self.ledger_api_handler.handle(incoming_message)

        # after
        assert self.game.phase == Phase.TOKENS_CREATED

        mock_logger.assert_any_call(logging.INFO, "tokens created.")

    def test_handle_transaction_receipt_callable_get_mint_batch_transaction_i(self):
        """Test the _handle_transaction_receipt method of the ledger_api handler where contract_api callable is GET_MINT_BATCH_TRANSACTION and all tokens are NOT minted."""
        # setup
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[:5],
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_messages[:4],
            ),
        )
        contract_api_dialogue = cast(
            ContractApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_messages[:4],
            ),
        )

        signing_dialogue.associated_contract_api_dialogue = contract_api_dialogue
        ledger_api_dialogue.associated_signing_dialogue = signing_dialogue

        contract_api_dialogue.callable = (
            ContractApiDialogue.Callable.GET_MINT_BATCH_TRANSACTION
        )

        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                transaction_receipt=self.transaction_receipt,
            ),
        )

        self.parameters.nb_completed_minting = 0
        self.game._registration._agent_addr_to_name = {
            "some_address_1": "some_name_1",
            "some_address_2": "some_name_2",
        }

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with patch.object(LedgerApis, "is_transaction_settled", return_value=True):
                self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(logging.INFO, "tokens minted.")
        assert self.parameters.nb_completed_minting == 1
        assert self.game.is_allowed_to_mint is True
        assert self.game.phase != Phase.TOKENS_MINTED

    def test_handle_transaction_receipt_callable_get_mint_batch_transaction_ii(self):
        """Test the _handle_transaction_receipt method of the ledger_api handler where contract_api callable is GET_MINT_BATCH_TRANSACTION and all tokens are minted."""
        # setup
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[:5],
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_messages[:4],
            ),
        )
        contract_api_dialogue = cast(
            ContractApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_messages[:4],
            ),
        )

        signing_dialogue.associated_contract_api_dialogue = contract_api_dialogue
        ledger_api_dialogue.associated_signing_dialogue = signing_dialogue

        contract_api_dialogue.callable = (
            ContractApiDialogue.Callable.GET_MINT_BATCH_TRANSACTION
        )

        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                transaction_receipt=self.transaction_receipt,
            ),
        )

        self.parameters.nb_completed_minting = 0
        self.game._registration._agent_addr_to_name = {"some_address": "some_name"}

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with patch.object(LedgerApis, "is_transaction_settled", return_value=True):
                self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(logging.INFO, "tokens minted.")
        assert self.parameters.nb_completed_minting == 1
        assert self.game.is_allowed_to_mint is True
        assert self.game.phase == Phase.TOKENS_MINTED
        mock_logger.assert_any_call(logging.INFO, "all tokens minted.")

    def test_handle_transaction_receipt_incorrect_callable(self):
        """Test the _handle_transaction_receipt method of the ledger_api handler where contract_api callable is incorrect."""
        # setup
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[:5],
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_messages[:4],
            ),
        )
        contract_api_dialogue = cast(
            ContractApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_messages[:4],
            ),
        )

        signing_dialogue.associated_contract_api_dialogue = contract_api_dialogue
        ledger_api_dialogue.associated_signing_dialogue = signing_dialogue

        contract_api_dialogue.callable = "some_incorrect_callable"

        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                transaction_receipt=self.transaction_receipt,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with patch.object(LedgerApis, "is_transaction_settled", return_value=True):
                self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(logging.ERROR, "unexpected transaction receipt!")

    def test_handle_error(self):
        """Test the _handle_error method of the ledger_api handler."""
        # setup
        ledger_api_dialogue = self.prepare_skill_dialogue(
            dialogues=self.ledger_api_dialogues,
            messages=self.list_of_ledger_api_messages[:1],
        )
        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.ERROR,
                code=1,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received ledger_api error message={incoming_message} in dialogue={ledger_api_dialogue}.",
        )

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the ledger_api handler."""
        # setup
        invalid_performative = LedgerApiMessage.Performative.GET_BALANCE
        incoming_message = self.build_incoming_message(
            message_type=LedgerApiMessage,
            dialogue_reference=("1", ""),
            performative=invalid_performative,
            ledger_id="some_ledger_id",
            address="some_address",
            to=self.skill.skill_context.agent_address,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle ledger_api message of performative={invalid_performative} in dialogue={self.ledger_api_dialogues.get_dialogue(incoming_message)}.",
        )

    def test_teardown(self):
        """Test the teardown method of the ledger_api handler."""
        assert self.ledger_api_handler.teardown() is None
        self.assert_quantity_in_outbox(0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains testing utilities."""
import asyncio
import logging
import sys
import time
from abc import ABC, abstractmethod
from threading import Timer
from typing import Dict, List, Optional

import docker
import pytest
import requests
from docker import DockerClient
from docker.models.containers import Container
from oef.agents import OEFAgent
from oef.core import AsyncioCore

from aea.exceptions import enforce


logger = logging.getLogger(__name__)


class DockerImage(ABC):
    """A class to wrap interatction with a Docker image."""

    def __init__(self, client: docker.DockerClient):
        """Initialize."""
        self._client = client

    def check_skip(self):
        """
        Check whether the test should be skipped.

        By default, nothing happens.
        """

    @property
    @abstractmethod
    def tag(self) -> str:
        """Return the tag of the image."""

    def stop_if_already_running(self):
        """Stop the running images with the same tag, if any."""
        client = docker.from_env()
        for container in client.containers.list():
            if self.tag in container.image.tags:
                logger.info(f"Stopping image {self.tag}...")
                container.stop()

    @abstractmethod
    def create(self) -> Container:
        """Instantiate the image in a container."""

    @abstractmethod
    def wait(self, max_attempts: int = 15, sleep_rate: float = 1.0) -> bool:
        """
        Wait until the image is running.

        :param max_attempts: max number of attempts.
        :param sleep_rate: the amount of time to sleep between different requests.
        :return: True if the wait was successful, False otherwise.
        """
        return True


class OEFHealthCheck(object):
    """A health check class."""

    def __init__(
        self,
        oef_addr: str,
        oef_port: int,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        """
        Initialize.

        :param oef_addr: IP address of the OEF node.
        :param oef_port: Port of the OEF node.
        """
        self.oef_addr = oef_addr
        self.oef_port = oef_port

        self._result = False
        self._stop = False
        self._core = AsyncioCore()
        self.agent = OEFAgent(
            "check", core=self._core, oef_addr=self.oef_addr, oef_port=self.oef_port
        )
        self.agent.on_connect_success = self.on_connect_ok
        self.agent.on_connection_terminated = self.on_connect_terminated
        self.agent.on_connect_failed = self.exception_handler

    def exception_handler(self, url=None, ex=None):
        """Handle exception during a connection attempt."""
        print("An error occurred. Exception: {}".format(ex))
        self._stop = True

    def on_connect_ok(self, url=None):
        """Handle a successful connection."""
        print("Connection OK!")
        self._result = True
        self._stop = True

    def on_connect_terminated(self, url=None):
        """Handle a connection failure."""
        print("Connection terminated.")
        self._stop = True

    def run(self) -> bool:
        """
        Run the check, asynchronously.

        :return: True if the check is successful, False otherwise.
        """
        self._result = False
        self._stop = False

        def stop_connection_attempt(self):
            if self.agent.state == "connecting":
                self.agent.state = "failed"

        t = Timer(1.5, stop_connection_attempt, args=(self,))

        try:
            print("Connecting to {}:{}...".format(self.oef_addr, self.oef_port))
            self._core.run_threaded()

            t.start()
            self._result = self.agent.connect()
            self._stop = True

            if self._result:
                print("Connection established. Tearing down connection...")
                self.agent.disconnect()
                t.cancel()
            else:
                print("A problem occurred. Exiting...")
            return self._result

        except Exception as e:
            print(str(e))
            return self._result
        finally:
            t.join(1.0)
            self.agent.stop()
            self.agent.disconnect()
            self._core.stop()


class OEFSearchDockerImage(DockerImage):
    """Wrapper to OEF Search Docker image."""

    def __init__(self, client: DockerClient, oef_addr: str, oef_port: int):
        """Initialize the OEF Search Docker image."""
        super().__init__(client)
        self._oef_addr = oef_addr
        self._oef_port = oef_port

    @property
    def tag(self) -> str:
        """Get the image tag."""
        return "fetchai/oef-search:0.7"

    def check_skip(self):
        """Check if the test should be skipped."""
        if sys.version_info < (3, 7):
            pytest.skip("Python version < 3.7 not supported by the OEF.")
            return

    def create(self) -> Container:
        """Create an instance of the OEF Search image."""
        from tests.conftest import ROOT_DIR  # pylint: disable

        logger.info(ROOT_DIR + "/tests/common/oef_search_pluto_scripts")
        ports = {
            "20000/tcp": ("0.0.0.0", 20000),  # nosec
            "30000/tcp": ("0.0.0.0", 30000),  # nosec
            "{}/tcp".format(self._oef_port): ("0.0.0.0", self._oef_port),  # nosec
        }
        volumes = {
            ROOT_DIR
            + "/tests/common/oef_search_pluto_scripts": {
                "bind": "/config",
                "mode": "rw",
            },
            ROOT_DIR + "/data/oef-logs": {"bind": "/logs", "mode": "rw"},
        }
        c = self._client.containers.run(
            self.tag,
            "/config/node_config.json",
            detach=True,
            ports=ports,
            volumes=volumes,
        )
        return c

    def wait(self, max_attempts: int = 15, sleep_rate: float = 1.0) -> bool:
        """Wait until the image is up."""
        success = False
        attempt = 0
        while not success and attempt < max_attempts:
            attempt += 1
            logger.info("Attempt {}...".format(attempt))
            oef_healthcheck = OEFHealthCheck("127.0.0.1", 10000)
            result = oef_healthcheck.run()
            if result:
                success = True
            else:
                logger.info(
                    "OEF not available yet - sleeping for {} second...".format(
                        sleep_rate
                    )
                )
                time.sleep(sleep_rate)

        return success


class GanacheDockerImage(DockerImage):
    """Wrapper to Ganache Docker image."""

    def __init__(
        self,
        client: DockerClient,
        addr: str,
        port: int,
        config: Optional[Dict] = None,
        gas_limit: int = 10000000000000,
    ):
        """
        Initialize the Ganache Docker image.

        :param client: the Docker client.
        :param addr: the address.
        :param port: the port.
        :param config: optional configuration to command line.
        """
        super().__init__(client)
        self._addr = addr
        self._port = port
        self._config = config or {}
        self._gas_limit = gas_limit

    @property
    def tag(self) -> str:
        """Get the image tag."""
        return "trufflesuite/ganache-cli:v6.12.1"

    def _make_ports(self) -> Dict:
        """Make ports dictionary for Docker."""
        return {f"{self._port}/tcp": ("0.0.0.0", self._port)}  # nosec

    def _build_command(self) -> List[str]:
        """Build command."""
        cmd = ["ganache-cli"]
        cmd += ["--gasLimit=" + str(self._gas_limit)]
        accounts_balances = self._config.get("accounts_balances", [])
        for account, balance in accounts_balances:
            cmd += [f"--account='{account},{balance}'"]
        return cmd

    def create(self) -> Container:
        """Create the container."""
        cmd = self._build_command()
        container = self._client.containers.run(
            self.tag, command=cmd, detach=True, ports=self._make_ports()
        )
        return container

    def wait(self, max_attempts: int = 15, sleep_rate: float = 1.0) -> bool:
        """Wait until the image is up."""
        request = dict(jsonrpc=2.0, method="web3_clientVersion", params=[], id=1)
        for i in range(max_attempts):
            try:
                response = requests.post(f"{self._addr}:{self._port}", json=request)
                enforce(response.status_code == 200, "")
                return True
            except Exception:
                logger.info(
                    "Attempt %s failed. Retrying in %s seconds...", i, sleep_rate
                )
                time.sleep(sleep_rate)
        return False



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: base.proto

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode("latin1"))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="base.proto",
    package="aea",
    syntax="proto3",
    serialized_options=None,
    serialized_pb=_b(
        '\n\nbase.proto\x12\x03\x61\x65\x61\x1a\x1cgoogle/protobuf/struct.proto"\x90\x01\n\x0f\x44ialogueMessage\x12\x12\n\nmessage_id\x18\x01 \x01(\x05\x12"\n\x1a\x64ialogue_starter_reference\x18\x02 \x01(\t\x12$\n\x1c\x64ialogue_responder_reference\x18\x03 \x01(\t\x12\x0e\n\x06target\x18\x04 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x05 \x01(\x0c"o\n\x07Message\x12\'\n\x04\x62ody\x18\x01 \x01(\x0b\x32\x17.google.protobuf.StructH\x00\x12\x30\n\x10\x64ialogue_message\x18\x02 \x01(\x0b\x32\x14.aea.DialogueMessageH\x00\x42\t\n\x07message"Y\n\x08\x45nvelope\x12\n\n\x02to\x18\x01 \x01(\t\x12\x0e\n\x06sender\x18\x02 \x01(\t\x12\x13\n\x0bprotocol_id\x18\x03 \x01(\t\x12\x0f\n\x07message\x18\x04 \x01(\x0c\x12\x0b\n\x03uri\x18\x05 \x01(\tb\x06proto3'
    ),
    dependencies=[google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,],
)


_DIALOGUEMESSAGE = _descriptor.Descriptor(
    name="DialogueMessage",
    full_name="aea.DialogueMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="message_id",
            full_name="aea.DialogueMessage.message_id",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="dialogue_starter_reference",
            full_name="aea.DialogueMessage.dialogue_starter_reference",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="dialogue_responder_reference",
            full_name="aea.DialogueMessage.dialogue_responder_reference",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="target",
            full_name="aea.DialogueMessage.target",
            index=3,
            number=4,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="aea.DialogueMessage.content",
            index=4,
            number=5,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=50,
    serialized_end=194,
)


_MESSAGE = _descriptor.Descriptor(
    name="Message",
    full_name="aea.Message",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="body",
            full_name="aea.Message.body",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="dialogue_message",
            full_name="aea.Message.dialogue_message",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="message",
            full_name="aea.Message.message",
            index=0,
            containing_type=None,
            fields=[],
        ),
    ],
    serialized_start=196,
    serialized_end=307,
)


_ENVELOPE = _descriptor.Descriptor(
    name="Envelope",
    full_name="aea.Envelope",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="to",
            full_name="aea.Envelope.to",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="sender",
            full_name="aea.Envelope.sender",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="protocol_id",
            full_name="aea.Envelope.protocol_id",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="message",
            full_name="aea.Envelope.message",
            index=3,
            number=4,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="uri",
            full_name="aea.Envelope.uri",
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=309,
    serialized_end=398,
)

_MESSAGE.fields_by_name[
    "body"
].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_MESSAGE.fields_by_name["dialogue_message"].message_type = _DIALOGUEMESSAGE
_MESSAGE.oneofs_by_name["message"].fields.append(_MESSAGE.fields_by_name["body"])
_MESSAGE.fields_by_name["body"].containing_oneof = _MESSAGE.oneofs_by_name["message"]
_MESSAGE.oneofs_by_name["message"].fields.append(
    _MESSAGE.fields_by_name["dialogue_message"]
)
_MESSAGE.fields_by_name["dialogue_message"].containing_oneof = _MESSAGE.oneofs_by_name[
    "message"
]
DESCRIPTOR.message_types_by_name["DialogueMessage"] = _DIALOGUEMESSAGE
DESCRIPTOR.message_types_by_name["Message"] = _MESSAGE
DESCRIPTOR.message_types_by_name["Envelope"] = _ENVELOPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DialogueMessage = _reflection.GeneratedProtocolMessageType(
    "DialogueMessage",
    (_message.Message,),
    dict(
        DESCRIPTOR=_DIALOGUEMESSAGE,
        __module__="base_pb2"
        # @@protoc_insertion_point(class_scope:aea.DialogueMessage)
    ),
)
_sym_db.RegisterMessage(DialogueMessage)

Message = _reflection.GeneratedProtocolMessageType(
    "Message",
    (_message.Message,),
    dict(
        DESCRIPTOR=_MESSAGE,
        __module__="base_pb2"
        # @@protoc_insertion_point(class_scope:aea.Message)
    ),
)
_sym_db.RegisterMessage(Message)

Envelope = _reflection.GeneratedProtocolMessageType(
    "Envelope",
    (_message.Message,),
    dict(
        DESCRIPTOR=_ENVELOPE,
        __module__="base_pb2"
        # @@protoc_insertion_point(class_scope:aea.Envelope)
    ),
)
_sym_db.RegisterMessage(Envelope)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the implementation of an autonomous economic agent (AEA)."""
import datetime
from asyncio import AbstractEventLoop
from logging import Logger
from multiprocessing.pool import AsyncResult
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    cast,
)

from aea.agent import Agent
from aea.agent_loop import AsyncAgentLoop, BaseAgentLoop, SyncAgentLoop
from aea.configurations.base import PublicId
from aea.configurations.constants import DEFAULT_SEARCH_SERVICE_ADDRESS
from aea.connections.base import Connection
from aea.context.base import AgentContext
from aea.crypto.ledger_apis import DEFAULT_CURRENCY_DENOMINATIONS
from aea.crypto.wallet import Wallet
from aea.decision_maker.base import DecisionMakerHandler
from aea.exceptions import AEAException, _StopRuntime
from aea.helpers.exception_policy import ExceptionPolicyEnum
from aea.helpers.logging import AgentLoggerAdapter, get_logger
from aea.helpers.temp_error_handler import ErrorHandler
from aea.identity.base import Identity
from aea.mail.base import Envelope
from aea.protocols.base import Message, Protocol
from aea.registries.filter import Filter
from aea.registries.resources import Resources
from aea.skills.base import Behaviour, Handler


class AEA(Agent):
    """This class implements an autonomous economic agent."""

    RUN_LOOPS: Dict[str, Type[BaseAgentLoop]] = {
        "async": AsyncAgentLoop,
        "sync": SyncAgentLoop,
    }
    DEFAULT_RUN_LOOP: str = "async"

    def __init__(
        self,
        identity: Identity,
        wallet: Wallet,
        resources: Resources,
        loop: Optional[AbstractEventLoop] = None,
        period: float = 0.05,
        execution_timeout: float = 0,
        max_reactions: int = 20,
        decision_maker_handler_class: Optional[Type[DecisionMakerHandler]] = None,
        skill_exception_policy: ExceptionPolicyEnum = ExceptionPolicyEnum.propagate,
        connection_exception_policy: ExceptionPolicyEnum = ExceptionPolicyEnum.propagate,
        loop_mode: Optional[str] = None,
        runtime_mode: Optional[str] = None,
        default_ledger: Optional[str] = None,
        currency_denominations: Optional[Dict[str, str]] = None,
        default_connection: Optional[PublicId] = None,
        default_routing: Optional[Dict[PublicId, PublicId]] = None,
        connection_ids: Optional[Collection[PublicId]] = None,
        search_service_address: str = DEFAULT_SEARCH_SERVICE_ADDRESS,
        **kwargs,
    ) -> None:
        """
        Instantiate the agent.

        :param identity: the identity of the agent
        :param wallet: the wallet of the agent.
        :param resources: the resources (protocols and skills) of the agent.
        :param loop: the event loop to run the connections.
        :param period: period to call agent's act
        :param execution_timeout: amount of time to limit single act/handle to execute.
        :param max_reactions: the processing rate of envelopes per tick (i.e. single loop).
        :param decision_maker_handler_class: the class implementing the decision maker handler to be used.
        :param skill_exception_policy: the skill exception policy enum
        :param loop_mode: loop_mode to choose agent run loop.
        :param runtime_mode: runtime mode (async, threaded) to run AEA in.
        :param default_ledger: default ledger id
        :param currency_denominations: mapping from ledger id to currency denomination
        :param default_connection: public id to the default connection
        :param default_routing: dictionary for default routing.
        :param connection_ids: active connection ids. Default: consider all the ones in the resources.
        :param search_service_address: the address of the search service used.
        :param kwargs: keyword arguments to be attached in the agent context namespace.

        :return: None
        """

        self._skills_exception_policy = skill_exception_policy
        self._connection_exception_policy = connection_exception_policy

        aea_logger = AgentLoggerAdapter(
            logger=get_logger(__name__, identity.name), agent_name=identity.name,
        )

        super().__init__(
            identity=identity,
            connections=[],
            loop=loop,
            period=period,
            loop_mode=loop_mode,
            runtime_mode=runtime_mode,
            logger=cast(Logger, aea_logger),
        )

        self.max_reactions = max_reactions

        if decision_maker_handler_class is None:
            from aea.decision_maker.default import (  # isort:skip  # pylint: disable=import-outside-toplevel
                DecisionMakerHandler as DefaultDecisionMakerHandler,
            )

            decision_maker_handler_class = DefaultDecisionMakerHandler
        decision_maker_handler = decision_maker_handler_class(
            identity=identity, wallet=wallet
        )
        self.runtime.set_decision_maker(decision_maker_handler)

        default_ledger_id = (
            default_ledger
            if default_ledger is not None
            else identity.default_address_key
        )
        currency_denominations = (
            currency_denominations
            if currency_denominations is not None
            else DEFAULT_CURRENCY_DENOMINATIONS
        )
        self._context = AgentContext(
            self.identity,
            self.runtime.multiplexer.connection_status,
            self.outbox,
            self.runtime.decision_maker.message_in_queue,
            decision_maker_handler.context,
            self.runtime.task_manager,
            default_ledger_id,
            currency_denominations,
            default_connection,
            default_routing if default_routing is not None else {},
            search_service_address,
            decision_maker_handler.self_address,
            **kwargs,
        )
        self._execution_timeout = execution_timeout
        self._connection_ids = connection_ids
        self._resources = resources
        self._filter = Filter(
            self.resources, self.runtime.decision_maker.message_out_queue
        )

        self._setup_loggers()

    @property
    def context(self) -> AgentContext:
        """Get (agent) context."""
        return self._context

    @property
    def resources(self) -> Resources:
        """Get resources."""
        return self._resources

    @resources.setter
    def resources(self, resources: "Resources") -> None:
        """Set resources."""
        self._resources = resources

    @property
    def filter(self) -> Filter:
        """Get the filter."""
        return self._filter

    @property
    def active_behaviours(self) -> List[Behaviour]:
        """Get all active behaviours to use in act."""
        return self.filter.get_active_behaviours()

    def setup(self) -> None:
        """
        Set up the agent.

        Performs the following:

        - loads the resources (unless in programmatic mode)
        - calls setup() on the resources

        :return: None
        """
        self.resources.setup()

    def act(self) -> None:
        """
        Perform actions.

        Calls act() of each active behaviour.

        :return: None
        """
        self.filter.handle_new_handlers_and_behaviours()

    @property
    def active_connections(self) -> List[Connection]:
        """Return list of active connections."""
        connections = self.resources.get_all_connections()
        if self._connection_ids is not None:
            connections = [
                c for c in connections if c.connection_id in self._connection_ids
            ]
        return connections

    def get_multiplexer_setup_options(self) -> Optional[Dict]:
        """
        Get options to pass to Multiplexer.setup.

        :return: dict of kwargs
        """
        return dict(
            connections=self.active_connections,
            default_routing=self.context.default_routing,
            default_connection=self.context.default_connection,
        )

    @staticmethod
    def _get_error_handler() -> Type[ErrorHandler]:
        """Get error handler."""
        handler = ErrorHandler
        return handler

    def _get_msg_and_handlers_for_envelope(
        self, envelope: Envelope
    ) -> Tuple[Optional[Message], List[Handler]]:
        """Get the msg and its handlers."""
        protocol = self.resources.get_protocol(envelope.protocol_id)

        msg, handlers = self._handle_decoding(envelope, protocol)

        return msg, handlers

    def _handle_decoding(
        self, envelope: Envelope, protocol: Optional[Protocol]
    ) -> Tuple[Optional[Message], List[Handler]]:

        handler = self._get_error_handler()

        if protocol is None:
            handler.send_unsupported_protocol(envelope, self.logger)
            return None, []  # Tuple[Optional[Message], List[Handler]]

        handlers = self.filter.get_active_handlers(
            protocol.public_id, envelope.skill_id
        )

        if len(handlers) == 0:
            handler.send_unsupported_skill(envelope, self.logger)
            return None, []

        if isinstance(envelope.message, Message):
            msg = envelope.message
            return msg, handlers
        try:
            msg = protocol.serializer.decode(envelope.message)
            msg.sender = envelope.sender
            msg.to = envelope.to
            return msg, handlers
        except Exception as e:  # pylint: disable=broad-except  # thats ok, because we send the decoding error back
            self.logger.warning("Decoding error. Exception: {}".format(str(e)))
            handler.send_decoding_error(envelope, self.logger)
            return None, []

    def handle_envelope(self, envelope: Envelope) -> None:
        """
        Handle an envelope.

        - fetching the protocol referenced by the envelope, and
        - returning an envelope to sender if the protocol is unsupported, using the error handler, or
        - returning an envelope to sender if there is a decoding error, using the error handler, or
        - returning an envelope to sender if no active handler is available for the specified protocol, using the error handler, or
        - handling the message recovered from the envelope with all active handlers for the specified protocol.

        :param envelope: the envelope to handle.
        :return: None
        """
        self.logger.debug("Handling envelope: {}".format(envelope))
        msg, handlers = self._get_msg_and_handlers_for_envelope(envelope)

        if msg is None:
            return

        for handler in handlers:
            handler.handle_wrapper(msg)

    def _setup_loggers(self):
        """Set up logger with agent name."""
        for element in [
            self.runtime.main_loop,
            self.runtime.multiplexer,
            self.runtime.task_manager,
            self.resources.component_registry,
            self.resources.behaviour_registry,
            self.resources.handler_registry,
            self.resources.model_registry,
        ]:
            element.logger = AgentLoggerAdapter(
                element.logger, agent_name=self._identity.name
            )

    def get_periodic_tasks(
        self,
    ) -> Dict[Callable, Tuple[float, Optional[datetime.datetime]]]:
        """
        Get all periodic tasks for agent.

        :return: dict of callable with period specified
        """
        tasks = super().get_periodic_tasks()
        tasks.update(self._get_behaviours_tasks())
        return tasks

    def _get_behaviours_tasks(
        self,
    ) -> Dict[Callable, Tuple[float, Optional[datetime.datetime]]]:
        """
        Get all periodic tasks for AEA behaviours.

        :return: dict of callable with period specified
        """
        tasks = {}

        for behaviour in self.active_behaviours:
            tasks[behaviour.act_wrapper] = (behaviour.tick_interval, behaviour.start_at)

        return tasks

    def get_message_handlers(self) -> List[Tuple[Callable[[Any], None], Callable]]:
        """
        Get handlers with message getters.

        :return: List of tuples of callables: handler and coroutine to get a message
        """
        return super().get_message_handlers() + [
            (self.filter.handle_internal_message, self.filter.get_internal_message,),
        ]

    def exception_handler(self, exception: Exception, function: Callable) -> bool:
        """
        Handle exception raised during agent main loop execution.

        :param exception: exception raised
        :param function: a callable exception raised in.

        :return: bool, propagate exception if True otherwise skip it.
        """
        # docstyle: ignore # noqa: E800
        def log_exception(e, fn, is_debug: bool = False):
            if is_debug:
                self.logger.debug(f"<{e}> raised during `{fn}`")
            else:
                self.logger.exception(f"<{e}> raised during `{fn}`")

        if self._skills_exception_policy == ExceptionPolicyEnum.propagate:
            log_exception(exception, function, is_debug=True)
            return True

        if self._skills_exception_policy == ExceptionPolicyEnum.stop_and_exit:
            log_exception(exception, function)
            raise _StopRuntime(
                AEAException(
                    f"AEA was terminated cause exception `{exception}` in skills {function}! Please check logs."
                )
            )

        if self._skills_exception_policy == ExceptionPolicyEnum.just_log:
            log_exception(exception, function)
            return False

        raise AEAException(
            f"Unsupported exception policy: {self._skills_exception_policy}"
        )

    def teardown(self) -> None:
        """
        Tear down the agent.

        Performs the following:

        - tears down the resources.

        :return: None
        """
        self.resources.teardown()

    def get_task_result(self, task_id: int) -> AsyncResult:
        """
        Get the result from a task.

        :return: async result for task_id
        """
        return self.runtime.task_manager.get_task_result(task_id)

    def enqueue_task(
        self, func: Callable, args: Sequence = (), kwds: Optional[Dict[str, Any]] = None
    ) -> int:
        """
        Enqueue a task with the task manager.

        :param func: the callable instance to be enqueued
        :param args: the positional arguments to be passed to the function.
        :param kwds: the keyword arguments to be passed to the function.
        :return the task id to get the the result.
        :raises ValueError: if the task manager is not running.
        """
        return self.runtime.task_manager.enqueue_task(func, args, kwds)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the dialogue classes of the tac control skill."""

from pathlib import Path
from typing import cast

from aea.test_tools.test_skill import BaseSkillTestCase, COUNTERPARTY_ADDRESS

from packages.fetchai.protocols.default.message import DefaultMessage
from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.protocols.tac.message import TacMessage
from packages.fetchai.skills.tac_control.dialogues import (
    DefaultDialogue,
    DefaultDialogues,
    OefSearchDialogue,
    OefSearchDialogues,
    TacDialogue,
    TacDialogues,
)

from tests.conftest import ROOT_DIR


class TestTacDialogues(BaseSkillTestCase):
    """Test dialogue classes of tac control."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_control")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.default_dialogues = cast(
            DefaultDialogues, cls._skill.skill_context.default_dialogues
        )
        cls.tac_dialogues = cast(TacDialogues, cls._skill.skill_context.tac_dialogues)
        cls.oef_search_dialogues = cast(
            OefSearchDialogues, cls._skill.skill_context.oef_search_dialogues
        )

    def test_default_dialogues(self):
        """Test the DefaultDialogues class."""
        _, dialogue = self.default_dialogues.create(
            counterparty=COUNTERPARTY_ADDRESS,
            performative=DefaultMessage.Performative.BYTES,
            content=b"some_content",
        )
        assert dialogue.role == DefaultDialogue.Role.AGENT
        assert dialogue.self_address == self.skill.skill_context.agent_address

    def test_oef_search_dialogues(self):
        """Test the OefSearchDialogues class."""
        _, dialogue = self.oef_search_dialogues.create(
            counterparty=COUNTERPARTY_ADDRESS,
            performative=OefSearchMessage.Performative.SEARCH_SERVICES,
            query="some_query",
        )
        assert dialogue.role == OefSearchDialogue.Role.AGENT
        assert dialogue.self_address == self.skill.skill_context.agent_address

    def test_tac_dialogues(self):
        """Test the TacDialogues class."""
        _, dialogue = self.tac_dialogues.create(
            counterparty=COUNTERPARTY_ADDRESS,
            performative=TacMessage.Performative.REGISTER,
            agent_name="some_agent_name",
        )
        assert dialogue.role == TacDialogue.Role.CONTROLLER
        assert dialogue.self_address == self.skill.skill_context.agent_address



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains a the registration db model."""

import logging
import os
import sqlite3
from typing import List, Tuple

from aea.skills.base import Model


_default_logger = logging.getLogger(
    "aea.packages.fetchai.skills.carpark_detection.detection_database"
)


class RegistrationDB(Model):
    """Communicate between the database and the python objects."""

    def __init__(self, **kwargs):
        """Initialise the Detection Database Communication class."""
        custom_path = kwargs.pop("custom_path", None)
        super().__init__(**kwargs)
        this_dir = os.getcwd()
        self.db_path = (
            os.path.join(this_dir, "registration.db")
            if custom_path is None
            else custom_path
        )
        if not os.path.exists(os.path.dirname(os.path.abspath(self.db_path))):
            raise ValueError(f"Path={self.db_path} not valid!")
        self._initialise_backend()

    def _initialise_backend(self) -> None:
        """Set up database and initialise the tables."""
        if os.path.isfile(self.db_path):
            return
        self._execute_single_sql(
            "CREATE TABLE IF NOT EXISTS registered_table (address TEXT, ethereum_address TEXT, "
            "ethereum_signature TEXT, fetchai_signature TEXT, "
            "developer_handle TEXT, tweet TEXT)"
        )

    def set_registered(
        self,
        address: str,
        ethereum_address: str,
        ethereum_signature: str,
        fetchai_signature: str,
        developer_handle: str,
        tweet: str,
    ):
        """Record a registration."""
        command = "INSERT OR REPLACE INTO registered_table(address, ethereum_address, ethereum_signature, fetchai_signature, developer_handle, tweet) values(?, ?, ?, ?, ?, ?)"
        variables = (
            address,
            ethereum_address,
            ethereum_signature,
            fetchai_signature,
            developer_handle,
            tweet,
        )
        self._execute_single_sql(command, variables)

    def is_registered(self, address: str) -> bool:
        """Check if an address is registered."""
        command = "SELECT * FROM registered_table WHERE address=?"
        variables = (address,)
        result = self._execute_single_sql(command, variables)
        return len(result) != 0

    def get_all_registered(self) -> List[str]:
        """Get all registered AW-1 AEAs."""
        command = "SELECT address FROM registered_table"
        variables = ()
        results = self._execute_single_sql(command, variables)
        registered = [result[0] for result in results]
        return registered

    def _execute_single_sql(
        self,
        command: str,
        variables: Tuple[str, ...] = (),
        print_exceptions: bool = True,
    ) -> List[Tuple[str, ...]]:
        """Query the database - all the other functions use this under the hood."""
        conn = None
        ret = []
        try:
            conn = sqlite3.connect(self.db_path, timeout=300)  # 5 mins
            c = conn.cursor()
            c.execute(command, variables)
            ret = c.fetchall()
            conn.commit()
        except Exception as e:  # pragma: nocover # pylint: disable=broad-except
            if print_exceptions:
                self.context.logger.warning(f"Exception in database: {e}")
        finally:
            if conn is not None:
                conn.close()

        return ret
